module gx_rst_ctrl (
	clk,
	reset,
	gxb_pwrdn_in,
	pll_locked,
	rx_freqlocked,
	gxb_powerdown,
	tx_digitalreset,
	rx_analogreset,
	rx_digitalreset
	);

	input		clk;
	input		reset;
	input		gxb_pwrdn_in;
	input		pll_locked;
	input		rx_freqlocked;
	output		gxb_powerdown;
	output		tx_digitalreset; 
	output		rx_analogreset;
	output		rx_digitalreset;

	reg			gxb_powerdown;
	reg			tx_digitalreset;
	reg			rx_analogreset;
	reg			rx_digitalreset;
	
	reg			gxb_pwrdn_in_sync;
	reg			gxb_pwrdn_in_sync_d1;
	reg			gxb_pwrdn_in_sync_d2;
	reg			gxb_pwrdn_in_sync_d3;
	reg			gxb_pwrdn_in_stable;
	reg			pll_locked_sync;
	reg			pll_locked_sync_d1;
	reg			pll_locked_sync_d2;
	reg			pll_locked_sync_d3;
	reg			pll_locked_stable;
	reg			rx_freqlocked_sync;
	reg			rx_freqlocked_sync_d1;
	reg			rx_freqlocked_sync_d2;
	reg			rx_freqlocked_sync_d3;
	reg			rx_freqlocked_stable;
	
	reg	[6:0]	count_gxb_pwrdn;
	reg			count_rx_analogreset;
	reg	[8:0]	count_rx_digitalreset;
	reg [2:0]	state;
	reg [2:0]	nextstate;
	
	// State Definitions
	parameter [2:0] stm_idle			= 3'b000;
	parameter [2:0] stm_gxb_pwrdn		= 3'b001;
	parameter [2:0] stm_pll_locked		= 3'b010;
	parameter [2:0] stm_tx_digitalreset	= 3'b011;
	parameter [2:0] stm_rx_analogreset	= 3'b100;
	parameter [2:0] stm_rx_freqlocked	= 3'b101;
	parameter [2:0] stm_rx_digitalreset	= 3'b110;
   
	altera_std_synchronizer #(1) gxb_pwrdn_in_altera_std_synchronizer (
		.clk ( clk ), 
		.reset_n ( ~reset ), 
		.din ( gxb_pwrdn_in ), 
		.dout ( gxb_pwrdn_in_sync )
	);
	
	always @ (posedge clk or posedge reset)
	begin
		if (reset == 1'b1) begin
			gxb_pwrdn_in_sync_d1 <= 1'b0;
			gxb_pwrdn_in_sync_d2 <= 1'b0;		
			gxb_pwrdn_in_sync_d3 <= 1'b0;
		end
		else begin
			gxb_pwrdn_in_sync_d1 <= gxb_pwrdn_in_sync;
			gxb_pwrdn_in_sync_d2 <= gxb_pwrdn_in_sync_d1;		
			gxb_pwrdn_in_sync_d3 <= gxb_pwrdn_in_sync_d2; 
		end
	end
	
	always @ (posedge clk or posedge reset)
	begin
		if (reset == 1'b1) begin
			gxb_pwrdn_in_stable <= 1'b0;
		end
		else begin
			gxb_pwrdn_in_stable <= gxb_pwrdn_in_sync & gxb_pwrdn_in_sync_d1 & gxb_pwrdn_in_sync_d2 & gxb_pwrdn_in_sync_d3;
		end
	end
	
	altera_std_synchronizer #(1) pll_locked_altera_std_synchronizer (
		.clk ( clk ), 
		.reset_n ( ~reset ), 
		.din ( pll_locked ), 
		.dout ( pll_locked_sync )
	);
	
	always @ (posedge clk or posedge reset)
	begin
		if (reset == 1'b1) begin
			pll_locked_sync_d1 <= 1'b0;
			pll_locked_sync_d2 <= 1'b0;		
			pll_locked_sync_d3 <= 1'b0;
		end
		else begin
			pll_locked_sync_d1 <= pll_locked_sync;
			pll_locked_sync_d2 <= pll_locked_sync_d1;		
			pll_locked_sync_d3 <= pll_locked_sync_d2; 
		end
	end
	
	always @ (posedge clk or posedge reset)
	begin
		if (reset == 1'b1) begin
			pll_locked_stable <= 1'b0;
		end
		else begin
			pll_locked_stable <= pll_locked_sync & pll_locked_sync_d1 & pll_locked_sync_d2 & pll_locked_sync_d3;
		end
	end
	
	altera_std_synchronizer #(1) rx_freqlocked_altera_std_synchronizer (
		.clk ( clk ), 
		.reset_n ( ~reset ), 
		.din ( rx_freqlocked ), 
		.dout ( rx_freqlocked_sync )
	);
	
	always @ (posedge clk or posedge reset)
	begin
		if (reset == 1'b1) begin
			rx_freqlocked_sync_d1 <= 1'b0;
			rx_freqlocked_sync_d2 <= 1'b0;		
			rx_freqlocked_sync_d3 <= 1'b0;
		end
		else begin
			rx_freqlocked_sync_d1 <= rx_freqlocked_sync;
			rx_freqlocked_sync_d2 <= rx_freqlocked_sync_d1;		
			rx_freqlocked_sync_d3 <= rx_freqlocked_sync_d2; 
		end
	end
	
	always @ (posedge clk or posedge reset)
	begin
		if (reset == 1'b1) begin
			rx_freqlocked_stable <= 1'b0;
		end
		else begin
			rx_freqlocked_stable <= rx_freqlocked_sync & rx_freqlocked_sync_d1 & rx_freqlocked_sync_d2 & rx_freqlocked_sync_d3;
		end
	end
	
	// FSM
	always @ (posedge clk or posedge reset)
	begin
		if (reset == 1'b1) begin
			state <= stm_idle;    
		end
		else begin
			state <= nextstate;   
		end
	end
   
	always @ * 
	begin
    case (state) 
	stm_idle:
		if (gxb_pwrdn_in_stable == 1'b1) begin
			nextstate = stm_gxb_pwrdn;
		end
		else begin
			nextstate = stm_idle;
		end
	stm_gxb_pwrdn:
		if (gxb_pwrdn_in_stable == 1'b0) begin
			nextstate = stm_pll_locked;
		end
		else begin
			nextstate = stm_gxb_pwrdn;
		end
	stm_pll_locked:
		if (gxb_pwrdn_in_stable == 1'b1) begin
			nextstate = stm_gxb_pwrdn;
		end
		else if (pll_locked_stable == 1'b1) begin
			nextstate = stm_tx_digitalreset;
		end
		else begin
			nextstate = stm_pll_locked;
		end
	stm_tx_digitalreset:
		if (gxb_pwrdn_in_stable == 1'b1) begin
			nextstate = stm_gxb_pwrdn;
		end
		else if (pll_locked_stable == 1'b0) begin
			nextstate = stm_pll_locked;
		end
		else if (count_rx_analogreset == 1'b1) begin
			nextstate = stm_rx_analogreset;
		end
		else begin
			nextstate = stm_tx_digitalreset; 
		end
	stm_rx_analogreset:
		if (gxb_pwrdn_in_stable == 1'b1) begin
			nextstate = stm_gxb_pwrdn;
		end
		else if (pll_locked_stable == 1'b0) begin
			nextstate = stm_pll_locked;
		end
		else if (rx_freqlocked_stable == 1'b1) begin
			nextstate = stm_rx_freqlocked;
		end
		else begin
			nextstate = stm_rx_analogreset;
		end
	stm_rx_freqlocked:
		if (gxb_pwrdn_in_stable == 1'b1) begin
			nextstate = stm_gxb_pwrdn;
		end
		else if (pll_locked_stable == 1'b0) begin
			nextstate = stm_pll_locked;
		end
		else if (rx_freqlocked_stable == 1'b0) begin
			nextstate = stm_rx_analogreset;
		end
		else if (count_rx_digitalreset == 9'd499) begin
			nextstate = stm_rx_digitalreset;
		end
		else begin
			nextstate = stm_rx_freqlocked; 
		end
	stm_rx_digitalreset:
		if (gxb_pwrdn_in_stable == 1'b1) begin
			nextstate = stm_gxb_pwrdn;
		end
		else if (pll_locked_stable == 1'b0) begin
			nextstate = stm_pll_locked;
		end
		else if (rx_freqlocked_stable == 1'b0) begin
			nextstate = stm_rx_analogreset;
		end
		else begin
			nextstate = stm_rx_digitalreset;
		end
	default: 
		begin
			nextstate = stm_idle;
		end
	endcase
	end
	
	always @ (posedge clk or posedge reset) 
	begin
		if (reset == 1'b1) begin
			gxb_powerdown <= 1'b0;
			tx_digitalreset <= 1'b1;
			rx_analogreset <= 1'b1;
			rx_digitalreset <= 1'b1; 
			count_rx_analogreset <= 1'b0;
			count_rx_digitalreset <= 9'd0;   
		end	
		else begin
		case (nextstate)
		stm_idle: 
			begin
				gxb_powerdown <= 1'b0;
				tx_digitalreset <= 1'b1;
				rx_analogreset <= 1'b1;
				rx_digitalreset <= 1'b1; 
				if (gxb_pwrdn_in == 1'b1) begin 
					count_gxb_pwrdn <= count_gxb_pwrdn + 7'd1;
				end
				else begin
					count_gxb_pwrdn <= 7'd0;
				end
				count_rx_analogreset <= 1'b0;
				count_rx_digitalreset <= 9'd0;
			end
		stm_gxb_pwrdn:
			begin
				gxb_powerdown <= 1'b1;
				tx_digitalreset <= 1'b1;
				rx_analogreset <= 1'b1;
				rx_digitalreset <= 1'b1; 
				count_gxb_pwrdn <= 7'd0;
				count_rx_analogreset <= 1'b0;
				count_rx_digitalreset <= 9'd0;
			end
		stm_pll_locked:
			begin
				gxb_powerdown <= 1'b0;
				tx_digitalreset <= 1'b1;
				rx_analogreset <= 1'b1;
				rx_digitalreset <= 1'b1;
				count_gxb_pwrdn <= 7'd0; 
				count_rx_analogreset <= 1'b0;
				count_rx_digitalreset <= 9'd0;
			end
		stm_tx_digitalreset:
			begin
				gxb_powerdown <= 1'b0;
				tx_digitalreset <= 1'b0;
				rx_analogreset <= 1'b1;
				rx_digitalreset <= 1'b1; 
				count_gxb_pwrdn <= 7'd0;
				count_rx_analogreset <= count_rx_analogreset + 1'b1; 
				count_rx_digitalreset <= 9'd0;
			end
		stm_rx_analogreset:
			begin
				gxb_powerdown <= 1'b0;
				tx_digitalreset <= 1'b0;
				rx_analogreset <= 1'b0;
				rx_digitalreset <= 1'b1; 
				count_gxb_pwrdn <= 7'd0;
				count_rx_analogreset <= 1'b0; 
				count_rx_digitalreset <= 9'd0;
			end
		stm_rx_freqlocked:
			begin
				gxb_powerdown <= 1'b0;
				tx_digitalreset <= 1'b0;
				rx_analogreset <= 1'b0;
				rx_digitalreset <= 1'b1; 
				count_gxb_pwrdn <= 7'd0;
				count_rx_analogreset <= 1'b0; 
				count_rx_digitalreset <= count_rx_digitalreset + 9'd1; 
			end
		stm_rx_digitalreset:
			begin
				gxb_powerdown <= 1'b0;
				tx_digitalreset <= 1'b0;
				rx_analogreset <= 1'b0;
				rx_digitalreset <= 1'b0; 
				if (gxb_pwrdn_in == 1'b1) begin 
					count_gxb_pwrdn <= count_gxb_pwrdn + 7'd1;
				end
				else begin
					count_gxb_pwrdn <= 7'd0;
				end
				count_rx_analogreset <= 1'b0; 
				count_rx_digitalreset <= 9'd0; 
			end
		default: 
			begin
				gxb_powerdown <= 1'b0;
				tx_digitalreset <= 1'b1;
				rx_analogreset <= 1'b1;
				rx_digitalreset <= 1'b1; 
				count_gxb_pwrdn <= 7'd0;
				count_rx_analogreset <= 1'b0;
				count_rx_digitalreset <= 9'd0;
			end
		endcase
		end
	end
	
endmodule