# choose simulation type: "functional" or "timing"
if {$argc >= 1} {
	set mode $1
} else {
	set mode "functional"
}
# choose timing model for timing simulation (use 0,1,2 for shortcuts)
# NOTE: altera transceivers can only be simulated for FPGA speed grade 8 !
if {$argc >= 2} {
	if {$2 == 0} {
		set sdo "lsc_8_1200mv_85c_vhd_slow.sdo"
	} elseif {$2 == 1} {
		set sdo "lsc_8_1200mv_0c_vhd_slow.sdo"
	} elseif {$2 == 2} {
		set sdo "lsc_min_1200mv_0c_vhd_fast.sdo"
	} else {
		set sdo $2
	}
} else {
	set sdo "lsc_8_1200mv_85c_vhd_slow.sdo"
}
# choose the order of initialization of LSC, LDC0, and LDC1
set LSCT  "1050ns"
set LDC0T "1050ns"
set LDC1T "1050ns"
if {$argc >= 3} {
	if {$3 == 0} {          # everything comes up at the same time
		set LSCT  "1050ns"
		set LDC0T "1050ns"
		set LDC1T "1050ns"
	} elseif {$3 == 1} {  	# FTK LDC stays permanently offline 
		set LSCT  "1050ns"
		set LDC0T "1050ns"
		set LDC1T "500000ns"
	} elseif {$3 == 2} {    # FTK LDC is online after DAQ link is fully up
		set LSCT  "1050ns"
		set LDC0T "1050ns"
		set LDC1T "12001ns"
	} elseif {$3 == 3} {    # FTK LDC is online before DAQ LDC is online
		set LSCT  "1050ns"
		set LDC0T "12001ns"
		set LDC1T "1050ns"
	} elseif {$3 == 4} {    # Both LDC's are online before LSC
		set LSCT  "12001ns"
		set LDC0T "1050ns"
		set LDC1T "1050ns"
	} else {
		set LSCT  "1050ns"
		set LDC0T "1050ns"
		set LDC1T "1050ns"	
	}
}

echo "PARAMETERS: " $mode $sdo

transcript on
quit -sim

vlib work
vlib gate_work
vlib hola_sim_lib
vcom -93 -work hola_sim_lib {hola_sim_lib.vhd}

# compile general altera megafunctions
vcom -93 -work work {AFIFOLSC.vhd}
vcom -93 -work work {AFIFOLDC.vhd}
vcom -93 -work work {ACLK50_FROM_100.vhd}
vcom -93 -work work {ACLK40_FROM_100.vhd}

# compile transceiver altera megafunctions
vcom -93 -work work {altgx_reco_single.vhd}
vcom -93 -work work {altgx_single.vhd}
vcom -93 -work work {altgx_reco_dual.vhd}
vcom -93 -work work {altgx_dual.vhd}
vcom -93 -work work {altgx_reco_mslave.vhd}
vcom -93 -work work {altgx_single_master.vhd}
vcom -93 -work work {altgx_single_slave.vhd}

# compile altgx wrappers
vcom -93 -work work {disparity_lookup.vhd}
vcom -93 -work work {tlk_wrap_tx.vhd}
vcom -93 -work work {tlk_wrap_tx_simple.vhd}
vcom -93 -work work {tlk_wrap_rx.vhd}
vcom -93 -work work {tlk_byte_order.vhd}
vcom -93 -work work {altgx_reset.vhd}
vcom -93 -work work {lsc_rres_simulator.vhd}
#vcom -93 -work work {altgx_tlk_dual.vhd}
#vcom -93 -work work {altgx_tlk_mslave.vhd}
#vcom -93 -work work {altgx_tlk_single.vhd}
vcom -93 -work work {altgx_tlk.vhd}

# compile hola and top-level wrapper
vcom -93 -work work {lrl_xoff_control.vhd}
vcom -93 -work work {core_holalsc.vhd}
vcom -93 -work work {core_holaldc.vhd}
vcom -93 -work work {lsc_top.vhd}
vcom -93 -work work {ldc_top.vhd}

# compile testbench
vcom -93 -work work {tb_dlink.vhd}

if {$mode == "functional"} {
	vcom -93 -work work {tb_dlink_functional.vhd}
	vsim -quiet -t ps work.cfg_tb_dlink_functional_altera -GPOWER_UP_RST_LSC_TIME=$LSCT -GPOWER_UP_RST_LDC0_TIME=$LDC0T -GPOWER_UP_RST_LDC1_TIME=$LDC1T
	
} elseif {$mode == "timing"} {
	# determine the name of sdo and vho timing files
	regsub -all .sdo $sdo "" sdobase
	regsub -all _vhd_ $sdobase _ vhobase
	echo "SDO/VHO FILES:" $sdobase $vhobase
	# compile gate-level model of lsc_top
	vcom -93 -work work simulation/modelsim/$vhobase.vho
	# recompile testbench using the new lsc_top
	vcom -93 -work work {tb_dlink_timing.vhd}
	vsim -quiet -t ps +transport_int_delays +transport_path_delays -sdftyp /HOLALSC1=simulation/modelsim/$sdobase.sdo -L altera_mf -L altera -L lpm -L sgate -L cycloneiv_hssi -L cycloneiv_pcie_hip -L cycloneiv -L gate_work -L work -voptargs="+acc" -GOUTFILE1NAME=out_dlinktm_fc_in.txt -GOUTFILE2DAQNAME=out_dlinktm_fc_out_daq.txt -GOUTFILE2FTKNAME=out_dlinktm_fc_out_ftk.txt -GOUTFILE3NAME=out_dlinktm_rc_in.txt -GOUTFILE4NAME=out_dlinktm_rc_out.txt -GPOWER_UP_RST_LSC_TIME=$LSCT -GPOWER_UP_RST_LDC0_TIME=$LDC0T -GPOWER_UP_RST_LDC1_TIME=$LDC1T work.cfg_tb_dlink_timing_altera
	source simulation/modelsim/lsc_dump_all_vcd_nodes.tcl
} else {
	puts "Please specify mode as either functional or timing"
	exit 0
}

view wave
delete wave *
add wave -noupdate -divider LSC_INPUTS
add wave -noupdate -in sim:/tb_dlink/holalsc1/*
add wave -noupdate -divider LSC_OUTPUTS
add wave -noupdate -out sim:/tb_dlink/holalsc1/*

add wave -noupdate -divider LDC0_INPUTS
add wave -noupdate -in sim:/tb_dlink/holaldc0/*
add wave -noupdate -divider LDC0_OUTPUTS
add wave -noupdate -out sim:/tb_dlink/holaldc0/*

add wave -noupdate -divider LDC1_INPUTS
add wave -noupdate -in sim:/tb_dlink/holaldc1/*
add wave -noupdate -divider LDC1_OUTPUTS
add wave -noupdate -out sim:/tb_dlink/holaldc1/*

add wave -noupdate -divider LDC_FSM0
add wave -noupdate {sim:/tb_dlink/holaldc0/altgx_wrap/rx0_ord/pres_state } 
add wave -noupdate {sim:/tb_dlink/holaldc0/ldc_inst/sim1/control_ldc_sim/curr_init } 
add wave -noupdate {sim:/tb_dlink/holaldc0/ldc_inst/sim1/control_ldc_sim/curr_ret_state } 
add wave -noupdate {sim:/tb_dlink/holaldc0/ldc_inst/fifoldc1/empty }
add wave -noupdate {sim:/tb_dlink/holaldc0/ldc_inst/fifoldc1/full } 
add wave -noupdate {sim:/tb_dlink/holaldc0/ldc_inst/testcheck1/curr_test_state }

add wave -noupdate -divider LDC_FSM1
add wave -noupdate {sim:/tb_dlink/holaldc1/altgx_wrap/rx0_ord/pres_state } 
add wave -noupdate {sim:/tb_dlink/holaldc1/ldc_inst/sim1/control_ldc_sim/curr_init } 
add wave -noupdate {sim:/tb_dlink/holaldc1/ldc_inst/sim1/control_ldc_sim/curr_ret_state } 
add wave -noupdate {sim:/tb_dlink/holaldc1/ldc_inst/fifoldc1/empty }
add wave -noupdate {sim:/tb_dlink/holaldc1/ldc_inst/fifoldc1/full } 
add wave -noupdate {sim:/tb_dlink/holaldc1/ldc_inst/testcheck1/curr_test_state }

if {$mode == "functional"} {
	add wave -noupdate -divider LSC_XOFF_ENA
	add wave -noupdate {sim:/tb_dlink/holalsc1/lsc_inst/LRLSIM/lrl_ctrl/pres_state } 
	add wave -noupdate -radix hexadecimal {sim:/tb_dlink/holalsc1/lsc_inst/LRLSIM/lrl_ctrl/lrl }
	add wave -noupdate -divider LSC_FSM
	add wave -noupdate {sim:/tb_dlink/holalsc1/PLL_CLK }
	add wave \
	{sim:/tb_dlink/holalsc1/lsc_inst/csim/control_sim/ldc_reset_i } \
	{sim:/tb_dlink/holalsc1/lsc_inst/csim/control_sim/ldown } 
	add wave -noupdate {sim:/tb_dlink/holalsc1/lsc_inst/csim/control_sim/curr_init } 
	add wave -noupdate {sim:/tb_dlink/holalsc1/lsc_inst/csim/control_sim/curr_res } 
	add wave -noupdate {sim:/tb_dlink/holalsc1/lsc_inst/framelsc1/curr_fr } 
	add wave -noupdate {sim:/tb_dlink/holalsc1/lsc_inst/split1/curr_split } 
}

configure wave -namecolwidth 270
run 75000ns
view wave
update
WaveRestoreZoom {0 ps} {75600 ns}
