# choose simulation type: "functional" or "timing"
if {$argc >= 1} {
	set mode $1
} else {
	set mode "functional"
}
# choose timing model for timing simulation (use 0,1,2 for shortcuts)
if {$argc >= 2} {
	if {$2 == 0} {
		set sdo "lsc_8_1200mv_85c_vhd_slow.sdo"
	} elseif {$2 == 1} {
		set sdo "lsc_8_1200mv_0c_vhd_slow.sdo"
	} elseif {$2 == 2} {
		set sdo "lsc_min_1200mv_0c_vhd_fast.sdo"
	} else {
		set sdo $2
	}
} else {
	set sdo "lsc_8_1200mv_85c_vhd_slow.sdo"
}
# choose whether to apply deposit to prescales (for speedup)
if {$argc >= 3} {
	set speedup $3
} else {
	set speedup "no"
}
echo "PARAMETERS: " $mode $sdo $speedup

transcript on
quit -sim

vlib work
vlib gate_work
vlib hola_sim_lib
vcom -93 -work hola_sim_lib {hola_sim_lib.vhd}

# compile general altera megafunctions
vcom -93 -work work {AFIFOLSC.vhd}
vcom -93 -work work {AFIFOLDC.vhd}
vcom -93 -work work {ACLK50_FROM_100.vhd}
vcom -93 -work work {ACLK40_FROM_100.vhd}

# compile transceiver altera megafunctions
vcom -93 -work work {altgx_reco_single.vhd}
vcom -93 -work work {altgx_single.vhd}
vcom -93 -work work {altgx_reco_dual.vhd}
vcom -93 -work work {altgx_dual.vhd}
vcom -93 -work work {altgx_reco_mslave.vhd}
vcom -93 -work work {altgx_single_master.vhd}
vcom -93 -work work {altgx_single_slave.vhd}

# compile altgx wrappers
vcom -93 -work work {disparity_lookup.vhd}
vcom -93 -work work {tlk_wrap_tx.vhd}
vcom -93 -work work {tlk_wrap_tx_simple.vhd}
vcom -93 -work work {tlk_wrap_rx.vhd}
vcom -93 -work work {tlk_byte_order.vhd}
vcom -93 -work work {altgx_reset.vhd}
vcom -93 -work work {lsc_rres_simulator.vhd}
#vcom -93 -work work {altgx_tlk_dual.vhd}
#vcom -93 -work work {altgx_tlk_mslave.vhd}
#vcom -93 -work work {altgx_tlk_single.vhd}
vcom -93 -work work {altgx_tlk.vhd}

# compile hola and top-level wrapper
vcom -93 -work work {lrl_xoff_control.vhd}
vcom -93 -work work {core_holalsc.vhd}
vcom -93 -work work {core_holaldc.vhd}
vcom -93 -work work {lsc_top.vhd}
vcom -93 -work work {ldc_top.vhd}

# compile testbench
vcom -93 -work work {tb_link.vhd}

if {$mode == "functional"} {
	vcom -93 -work work {tb_link_functional.vhd}
	vsim -quiet -t ps work.cfg_tb_link_functional_altera
	if {$speedup == "yes"} {
		force -deposit sim:/tb_link/holaldc1/ldc_inst/sim1/control_ldc_sim/prescale 000001 0 -cancel 100000
		force -deposit sim:/tb_link/holalsc1/lsc_inst/csim/control_sim/prescale 000001 0 -cancel 100000
	}
} elseif {$mode == "timing"} {
	# determine the name of sdo and vho timing files
	regsub -all .sdo $sdo "" sdobase
	regsub -all _vhd_ $sdobase _ vhobase
	echo "SDO/VHO FILES:" $sdobase $vhobase
	# compile gate-level model of lsc_top
	vcom -93 -work work simulation/modelsim/$vhobase.vho
	# recompile testbench using the new lsc_top
	vcom -93 -work work {tb_link_timing.vhd}
	vsim -quiet -t ps +transport_int_delays +transport_path_delays -sdftyp /HOLALSC1=simulation/modelsim/$sdobase.sdo -L altera_mf -L altera -L lpm -L sgate -L cycloneiv_hssi -L cycloneiv_pcie_hip -L cycloneiv -L gate_work -L work -voptargs="+acc" -GOUTFILE1NAME=out_linktm_fc_in.txt -GOUTFILE2NAME=out_linktm_fc_out.txt -GOUTFILE3NAME=out_linktm_rc_in.txt -GOUTFILE4NAME=out_linktm_rc_out.txt work.cfg_tb_link_timing_altera
	source simulation/modelsim/lsc_dump_all_vcd_nodes.tcl
} else {
	puts "Please specify mode as either functional or timing"
	exit 0
}

view wave
delete wave *
add wave -noupdate -divider LSC_INPUTS
add wave -noupdate -in sim:/tb_link/holalsc1/*
add wave -noupdate -divider LSC_OUTPUTS
add wave -noupdate -out sim:/tb_link/holalsc1/*
if {$mode == "functional"} {
	add wave -noupdate -divider LSC_FSM
	add wave -noupdate \
	{sim:/tb_link/holalsc1/altgx_wrap/tx_rst/pres_state } \
	{sim:/tb_link/holalsc1/altgx_wrap/rstsim/rx0_rst/pres_state } \
	{sim:/tb_link/holalsc1/altgx_wrap/rx0_ord/pres_state } \
	{sim:/tb_link/holalsc1/lsc_inst/csim/control_sim/curr_init } 
}
add wave -noupdate -divider LDC_INPUTS
add wave -noupdate -in sim:/tb_link/holaldc1/*
add wave -noupdate -divider LDC_OUTPUTS
add wave -noupdate -out sim:/tb_link/holaldc1/*
add wave -noupdate -divider LDC_FSM
add wave -noupdate {sim:/tb_link/holaldc1/altgx_wrap/rx_ord/pres_state } 
add wave -noupdate {sim:/tb_link/holaldc1/ldc_inst/sim1/control_ldc_sim/curr_init } 
add wave -noupdate {sim:/tb_link/holaldc1/ldc_inst/sim1/control_ldc_sim/curr_ret_state } 

configure wave -namecolwidth 270
run 75000ns
view wave
update
WaveRestoreZoom {0 ps} {75600 ns}
