# choose simulation type: "functional" or "timing"
if {$argc >= 1} {
	set mode $1
} else {
	set mode "functional"
}
# choose timing model for timing simulation
if {$argc >= 2} {
	set sdo $2
} else {
	set sdo "lsc_8_1200mv_85c_vhd_slow.sdo"
}
# enable debug waveforms by default (just keeps sending idles)
if {$argc >= 3} {
	set debug $3
} else {
	set debug 1
}

echo "PARAMETERS: " $mode $sdo

transcript on
quit -sim

vlib work
vlib gate_work

# compile altera components
vcom -93 -work work {altgx_reco_dual.vhd}
vcom -93 -work work {altgx_dual.vhd}
vcom -93 -work work {altgx_reco_single.vhd}
vcom -93 -work work {altgx_single.vhd}

# compile altgx wrappers
vcom -93 -work work {disparity_lookup.vhd}
vcom -93 -work work {tlk_wrap_tx.vhd}
vcom -93 -work work {tlk_wrap_rx.vhd}
vcom -93 -work work {tlk_byte_order.vhd}
vcom -93 -work work {altgx_reset.vhd}
vcom -93 -work work {altgx_tlk_dual.vhd}
vcom -93 -work work {altgx_tlk_single.vhd}

vcom -93 -work work {tb_tgxds.vhd}

if {$mode == "functional"} {
	vsim -quiet -t ps work.tb_tgxds(structure)
} elseif {$mode == "timing"} {
	# determine the name of sdo and vho timing files
	regsub -all .sdo $sdo "" sdobase
	regsub -all _vhd_ $sdobase _ vhobase
	echo "SDO/VHO FILES:" $sdobase $vhobase
	# compile gate-level model of lsc_top
	vcom -93 -work work simulation/modelsim/$vhobase.vho
	# recompile testbench using the new lsc_top
	vcom -93 -work work {tb_tgxds_timing.vhd}
	vsim -quiet -t ps +transport_int_delays +transport_path_delays -sdftyp /dual=simulation/modelsim/$sdobase.sdo -L altera_mf -L altera -L lpm -L sgate -L cycloneiv_hssi -L cycloneiv_pcie_hip -L cycloneiv -L gate_work -L work -voptargs="+acc" work.cfg_tb_tgxds_timing_altera
	#source simulation/modelsim/lsc_dump_all_vcd_nodes.tcl
} else {
	puts "Please specify mode as either functional or timing"
	exit 0
}

delete wave *

# general transmitter ports
add wave -noupdate \
{sim:/tb_tgxds/slowclk } \
{sim:/tb_tgxds/refclk } \
{sim:/tb_tgxds/power_up_rst_n0 } \
{sim:/tb_tgxds/power_up_rst_n1 }

add wave -noupdate -divider gx_status
add wave -noupdate {sim:/tb_tgxds/debug_bits }
if {$mode == "functional"} {
	add wave -noupdate \
	{sim:/tb_tgxds/dual/tx_rst/pres_state } \
	{sim:/tb_tgxds/dual/rstsim/rx0_rst/pres_state } \
	{sim:/tb_tgxds/dual/rx0_ord/pres_state } \
	{sim:/tb_tgxds/dual/rstsim/rx1_rst/pres_state } \
	{sim:/tb_tgxds/dual/rx1_ord/pres_state }
}

add wave -noupdate -divider tx
add wave -noupdate -label clk {sim:/tb_tgxds/TX_CLK}
add wave -noupdate -radix hexadecimal -label txd {sim:/tb_tgxds/txd }
add wave -noupdate -label tx_er {sim:/tb_tgxds/tx_er }
add wave -noupdate -label tx_en {sim:/tb_tgxds/tx_en }
if {$mode == "functional"} {
	add wave -noupdate -radix hexadecimal -label txsent \
	{sim:/tb_tgxds/dual/gx_inst/altgx_dual_alt_c3gxb_component/transmit_pma0/datain } 
}

add wave -noupdate -divider rx_daq
if {$mode == "functional"} {
	add wave -noupdate -radix hexadecimal -label TSRC {sim:/tb_tgxds/single0/gx_inst/altgx_single_alt_c3gxb_component/transmit_pma0/datain } 
}
add wave -noupdate -label rx_clkout {sim:/tb_tgxds/RX_CLK0 }
add wave -noupdate -radix hexadecimal -label RD {sim:/tb_tgxds/RXD0 }
add wave -noupdate -label RX_ER {sim:/tb_tgxds/RX_ER0 }
add wave -noupdate -label RX_DV {sim:/tb_tgxds/RX_DV0 }

add wave -noupdate -divider rx_ftk
if {$mode == "functional"} {
	add wave -noupdate -radix hexadecimal -label TSRC {sim:/tb_tgxds/single1/gx_inst/altgx_single_alt_c3gxb_component/transmit_pma0/datain } 
}
add wave -noupdate -label rx_clkout {sim:/tb_tgxds/RX_CLK1 }
add wave -noupdate -radix hexadecimal -label RD {sim:/tb_tgxds/RXD1 }
add wave -noupdate -label RX_ER {sim:/tb_tgxds/RX_ER1 }
add wave -noupdate -label RX_DV {sim:/tb_tgxds/RX_DV1 }

if {$mode == "functional"} {
	add wave -noupdate -divider rx_debug
	add wave -noupdate -label rx_freqlocked {sim:/tb_tgxds/dual/gx_inst/rx_freqlocked }
	add wave -noupdate -radix hexadecimal -label rxreceived {sim:/tb_tgxds/dual/gx_inst/altgx_dual_alt_c3gxb_component/receive_pma0/rxdata }
	add wave -noupdate -radix hexadecimal -label rx_dataout {sim:/tb_tgxds/dual/gx_inst/rx_dataout }
	add wave -noupdate -label rx_ctrldetect {sim:/tb_tgxds/dual/gx_inst/rx_ctrldetect }
	add wave -noupdate \
	{sim:/tb_tgxds/dual/gx_inst/rx_errdetect } \
	{sim:/tb_tgxds/dual/gx_inst/rx_disperr } \
	{sim:/tb_tgxds/dual/gx_inst/rx_runningdisp } \
	{sim:/tb_tgxds/dual/gx_inst/rx_patterndetect } \
	{sim:/tb_tgxds/dual/gx_inst/rx_syncstatus }
}


# clocks
force -freeze /tb_tgxds/REFCLK  0 0ns, 1 5ns -repeat 10ns
force -freeze /tb_tgxds/SLOWCLK  0 0ns, 1 10ns -repeat 20ns

if {$debug == 1} {
	# just have a continuous stream idles for debugging rx locking
	force -freeze /tb_tgxds/POWER_UP_RST_N0 0 0ns, 1 1100ns
	force -freeze /tb_tgxds/POWER_UP_RST_N1 0 0ns, 1 1100ns
	force -freeze /tb_tgxds/tx_en       1 0ns, 0 700ns,  1 12010ns
	force -freeze /tb_tgxds/tx_er       1 0ns, 0 700ns,  0 12010ns
	force -freeze /tb_tgxds/txd   16#FFFF 0ns,     16#ABCD 12010ns
} else {
	# transmit (first idle block to synchronize with DAQ, second - with FTK) - reset in first 10 ns!
	force -freeze /tb_tgxds/POWER_UP_RST_N0 0 0ns, 1 10ns
	force -freeze /tb_tgxds/POWER_UP_RST_N1 0 0ns, 1 5ns
	#                                          sync daq          data       data         sync ftk    data      carrier    error      data
	force -freeze /tb_tgxds/tx_en       1 0ns, 0 700ns, 1 740ns, 1 750ns,       1 1010ns, 0 1200ns, 1 1250ns, 0 1400ns, 1 1700ns, 1 1790ns
	force -freeze /tb_tgxds/tx_er       1 0ns, 0 700ns, 1 740ns, 0 750ns,       0 1010ns, 0 1200ns, 0 1250ns, 1 1400ns, 1 1700ns, 0 1790ns
	force -freeze /tb_tgxds/txd   16#FFFF 0ns,             16#0102 750ns, 16#ABCD 1010ns,     16#CDEF 1250ns,               16#CDBA 1790ns,16#CDBB 1800ns,16#CDBC 1810ns,16#CDBD 1820ns
}

run 15000ns
view wave
update
