#include <stdio.h>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

#define BOF_WORD BOF0AAA0
#define EOF_WORD EOF0AAA0

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 
#define FILAR_CHANNEL_0 0x1
#define FILAR_CHANNEL_1 0x2
#define FILAR_CHANNEL_MASK_0 0x1
#define FILAR_CHANNEL_MASK_1 0x2

// first FILAR channel:
#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

#define rdtscl(low) \
__asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

int timing() {
    static __u32 told;
    __u32 tnew;
    int result;
    rdtscl(tnew);
    result = tstamp_to_us(told,tnew);
    told = tnew;
    return result;
}

int str_to_int(string v) {
  int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

void receive(unsigned int FILAR_CHANNEL) {
    // use static array data[] for filar receive buffer:
    unsigned data[100000];
    // other variables
    int filsize = 0;
    int filerr(0);

    while (!filsize) {
      filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
    }

    cerr << "FILAR: received " << filsize 
	 << " words; error=" << hex << filerr << dec << endl;
    if(filerr!=0) {
      cerr << "ERROR: major frame error" << endl;
      exit(0);
    }
    for(int z=0; z<filsize; z++) {
      if(z>1 && z<filsize-2) continue; // for brevity
        cerr << "FILAR: data[" << z << "] = " 
            << hex << data[z] << dec << endl;
    }
}

int main( int argc, char* argv[]) {
  // length of frame that is sent through SOLAR
  int FRAMELEN(254);
  int NFRAMES(4);
  bool FTK_XOFF_ENA(false);
  int result;
  if(argc<2) {
    cout << "USAGE: " << argv[0] << " FRAMELEN NFRAMES FTK_XOFF_ENA" << endl;
    cout << "EXAMPLE: " << argv[0] << " 254 17 1" << endl;
    exit(0);
  }
  if(argc>=2) FRAMELEN = str_to_int(argv[1]);
  if(argc>=3) NFRAMES = str_to_int(argv[2]);
  if(argc>=4) FTK_XOFF_ENA = str_to_int(argv[3]);
  cout <<"FRAMELEN=" << FRAMELEN 
       << "; NFRAMES=" << NFRAMES
       << "; FTK_XOFF_ENA=" << FTK_XOFF_ENA
       << endl;
  
  unsigned int *dataptr_solar;

  cerr << "SOLAR: before solar_setup" << endl;
  solar_setup(0);
  solar_slidas_off(0);
  cerr << "SOLAR: after solar_setup" << endl;
  dataptr_solar = solar_getbuffer(0);
  cerr << "SOLAR: retrieved solar buffer!"  << endl;
 
  // Set up two FILAR channels
  cerr << "FILAR: filar_setup with channel mask: " 
       << (FILAR_CHANNEL_MASK_BOTH) << endl;
  int setup_status = filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK_BOTH);
  if(setup_status==-1) {
    cerr << "ERROR: failed to bring up FILAR" << endl;
    exit(0);
  }

  if(FTK_XOFF_ENA) {
    cerr << "SOLAR->FILAR: enabling FTK_XOFF_ENA register "
	 <<  "through the return line" << endl;
    filar_flowctrlenable(0, FILAR_CHANNEL_1);
  }
  
  // arm a few REQ FIFOs before going to event loop
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_1, FILAR_ARM);
  int req_armed[] = {FILAR_ARM,FILAR_ARM};
  int totr[] = {1,1}; // start count with one
  
  cerr << "STATS: done initialization" << endl;

  // prepare some dummy data
  int sent=0;
  for(int il=0;il<FRAMELEN;il++) {*(dataptr_solar+il)=0xaaaaaaaa+il;}

  cerr << "SOLAR: sending " << FRAMELEN << " words" << endl;
  solar_send(FRAMELEN,0); sent++;
  cerr << "SOLAR: sent " << FRAMELEN << " words" << endl;
  cerr << "---> sent count = " << sent << endl;
  
  // now try to trigger XOFF from FILAR end
  cerr << "SOLAR: sending data train of " << NFRAMES-1 << " frames" << endl;
  for(int il=0;il<NFRAMES-1;il++) {
    *(dataptr_solar+FRAMELEN-1) = 0xabcdaaaa+il;
    solar_send(FRAMELEN,0); sent++;
    cerr << "---> sent count = " << sent << endl;
  }

  // Continuous loop to control data reception
  cerr << "FYI, sent a total of " << sent << " frames ("
       << sent*FRAMELEN << ") words" << endl;
  while(true) {
    unsigned int chanmask;
    int nreceive;
    while(true) {
      cerr << "Enter CHANMASK[1..3] NBUFFERS[1..30]: ";
      cin >> chanmask >> nreceive;
      cerr << "User input: " << chanmask << " " << nreceive << endl;
      if( (chanmask>0 && chanmask<=3) && (nreceive>0 && nreceive<=30) ) {
	break;
      }
    }
    for(int ir=0;ir<nreceive;ir++) {
      if(chanmask & FILAR_CHANNEL_0) {
	cerr << "=========== " << totr[0] << " - CHANNEL" 
	     << FILAR_CHANNEL_0 << "===========" << endl;
	receive(FILAR_CHANNEL_0);
	req_armed[0]--;
	totr[0]++;
      }
      if(chanmask & FILAR_CHANNEL_1) {
	cerr << "=========== " << totr[1] << " - CHANNEL" 
	     << FILAR_CHANNEL_1 << "===========" << endl;
	receive(FILAR_CHANNEL_1);
	req_armed[1]--;
	totr[1]++;
      }
      // time to re-arm the REQ FIFO
      for(unsigned int ichan=0; ichan<=1; ichan++) {
	if(req_armed[ichan]==0) {
	  cerr << "INFO: Arming " << ichan << " with " << FILAR_ARM << " REQ FIFO buffers..." << endl;
	  filar_setreq(FILAR_SLOT, ichan==0 ? FILAR_CHANNEL_0 : FILAR_CHANNEL_1, FILAR_ARM);
	  sleep(0.1);
	  req_armed[ichan] = FILAR_ARM;
	  cerr << "Armed!" << endl;
	}
      }
    }
  }
}
