#include <stdio.h>
#include <cstdlib>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

#define BOF_WORD BOF0AAA0
#define EOF_WORD EOF0AAA0

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 10
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 

// first FILAR channel:
#if 0
#define FILAR_CHANNEL_MASK 0x1
#define FILAR_CHANNEL_0 0x1
#else
// OR: second FILAR channel:
#define FILAR_CHANNEL_MASK 0x2
#define FILAR_CHANNEL_0 0x2
#define FILAR_CHANNEL_1 0x1
#endif
#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

#define rdtscl(low) \
__asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

int timing() {
    static __u32 told;
    __u32 tnew;
    int result;
    rdtscl(tnew);
    result = tstamp_to_us(told,tnew);
    told = tnew;
    return result;
}

int str_to_int(string v) {
  int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
int receive(unsigned int FILAR_CHANNEL, int nsent, unsigned int *dataptr_solar)
{
  // use static array data[] for filar receive buffer:
  unsigned data[10000];
  // other variables
  int filsize = 0;
  int filerr(0);
  int errcnt = 0;
  while (!filsize) {
    filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
  }
  if(filerr!=0 || filsize != nsent) {
    cerr << "FILAR ERROR: received " << filsize << " / " << nsent
	 << " words; error=" << hex << filerr << dec << endl;
    return -1;
  }
  
  for(int z=0; z<MIN(nsent,filsize); z++) {
    // break; // FIXME - has minimal effect on speed (4%)
    if(data[z] != dataptr_solar[z]) {
      errcnt+=1;
      std::cerr << "FILAR ERROR: sent [" << hex << dataptr_solar[z] 
		<< "], received [" << data[z] << "]" << dec
		<< std::endl;
    }
  }
  return errcnt;
}

int main( int argc, char* argv[]) {
  std::cout << "Max random number bound: "
	    << RAND_MAX << std::endl;
  // length of frame that is sent through SOLAR
  long int NFRAMES(3000);
  long int FRAMELEN(254);
  if(argc==1) {
    cout << "USAGE: " << argv[0] << "NFRAMES FRAMELEN" << endl;
  }
  if(argc>=2) NFRAMES = str_to_int(argv[1]);
  if(argc>=3) FRAMELEN = str_to_int(argv[2]);
  cout << "NFRAMES=" << NFRAMES << "; FRAMELEN=" << FRAMELEN << endl;
  
  unsigned int *dataptr_solar;

  cerr << "SOLAR: before solar_setup" << endl;
  solar_setup(0);
  cerr << "SOLAR[" << "]: after solar_setup" << endl;
  dataptr_solar = solar_getbuffer(0);
  cerr << "SOLAR: retrieved solar buffer!"  << endl;
 
  // We have a faulty SOLAR that doesn't always come up
  // The way to detect failure is by looking at FILAR LDOWN
  // If detected, we re-reset SOLAR and try again.
  int linkup=0;
  for(int itry=0; itry<3; itry++) {
    cerr << "FILAR: before filar_setup (attempt " << itry << ")" << endl;   
    if( filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK_BOTH) == -1) {
      solar_setup(0);
      cerr << "SOLAR[" << "]: after solar_setup" << endl;
      dataptr_solar = solar_getbuffer(0);
      }
    else {
      cerr << "FILAR[" << "] after filar_setup" << endl;
      linkup=1;
      break;
    }
  }
  if(linkup==0) {
    cerr << "ERROR: unable to bring up FILAR" << endl;
    exit(0);
  }
  // arm a few REQ FIFOs before going to event loop
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_1, FILAR_ARM);
  int req_armed = FILAR_ARM;
  
  cerr << "Running tests. You won't see any output until it's done..." << endl;

  // prepare some dummy data
  long int iframe = NFRAMES;
  time_t beg_time = time(NULL);
  long int errcnt_word = 0;
  long int errcnt_frame = 0;
  while(iframe-- > 0) {
    // generating data to send - skipping this gives a 25% speedup
    for(int il=0; il<FRAMELEN;il++) *(dataptr_solar+il)=rand();
    // sending data
    solar_send(FRAMELEN,0);
    // receiving
    for(int ir=0;ir<=1;ir++) {
      int status = receive(ir+1,FRAMELEN,dataptr_solar);
      if(status == -1) errcnt_frame++; else errcnt_word+=status;
      req_armed--;
      // time to re-arm the REQ FIFO
      if(req_armed<=1) {
	filar_setreq(FILAR_SLOT, ir+1, FILAR_ARM);
	//filar_setreq(FILAR_SLOT, 2, FILAR_ARM);
	//sleep(0.1);
	req_armed = FILAR_ARM;
      }
    }
  }
  time_t end_time = time(NULL);
  cout << "=========================================" << endl;
  cout << "BAD: frames = " << errcnt_frame 
       << ", words = " << errcnt_word << endl;
  const double nwords = NFRAMES*FRAMELEN;
  const double nbits = NFRAMES*FRAMELEN*32.0;
  cout << "TOTAL: frames = " << NFRAMES 
       << ", words = " << nwords
       << ", bits = " << nbits << endl;
  const double dtime = end_time - beg_time;
  cout << "TOTAL TIME (sec) = " << dtime << endl;
  if(dtime>0) {
    const double bps = 1.0*nbits/dtime;
    cout << "Bit/sec = " << bps << endl;
    cout << "MBit/sec = " << bps/(1024*1024) << endl;
    const double tproj = 1e15/bps;
    cout << "Time to test 1e15 bits: " << tproj << " sec, " << tproj/(60.0*60.0*24) << " days" << endl;
  }
  cout << "=========================================" << endl;
}


