#include <stdio.h>
#include <cstdlib>
#include <cassert>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include <sys/time.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

//#define ADD_SLEEP

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 
#define FILAR_CHANNEL_0 0x1
#define FILAR_CHANNEL_1 0x2
#define FILAR_CHANNEL_MASK_0 0x1
#define FILAR_CHANNEL_MASK_1 0x2

#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

#define rdtscl(low) \
__asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

int timing() {
    static __u32 told;
    __u32 tnew;
    int result;
    rdtscl(tnew);
    result = tstamp_to_us(told,tnew);
    told = tnew;
    return result;
}

unsigned int str_to_uint(string v) {
  unsigned int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

int slidas_random_old(int reset)
{
  static unsigned reg_word = 0x524948e0; /* Random number seed */
  unsigned output_word;
  int i, j, xor_result;
  static int swap_map[32] = { 19,  0,  9, 20,  5, 31, 24, 22, 
			      1,  7, 27,  8, 17, 29, 23, 11, 
			      18, 21, 25,  2,  6, 13,  3, 10, 
			      30, 28, 14, 12,  4,  0, 26, 15};
  if (reset) reg_word = 0x524948e0;				   
  xor_result = 
    ((((reg_word & 0x80000000) && 1) ^     /* Bit 31 */
      ((reg_word & 0x20000000) && 1) ^     /* Bit 29 */
      ((reg_word & 0x40000) && 1)    ^     /* Bit 18 */
      ((reg_word & 0x20) && 1)) == 1);     /* Bit  5 */
  output_word = 0;
  for(j=0; j<32; j++) {
    if (j != 29) {
      if (j == 16)		 	    
	output_word ^= (xor_result << 16); 
      if (reg_word & (0x1 << j)) 	
	output_word ^= (0x1 << swap_map[j]);
    }
  }
  reg_word = output_word;
  return output_word; 
}

#define SEED 0x524948e0 /* Random number seed */

typedef union {
    unsigned int ival;
    unsigned short int sval[2];
} WORD;
    
static WORD reg_word;

unsigned int htab[65536], ltab[65536];

unsigned int slidas_random_next(unsigned int val)
{
    unsigned output_word, xor_result;
    int j;
    static int swap_map[32] = { 19,  0,  9, 20,  5, 31, 24, 22, 
				 1,  7, 27,  8, 17, 29, 23, 11, 
				18, 21, 25,  2,  6, 13,  3, 10, 
				30, 28, 14, 12,  4,  0, 26, 15 };
    

    xor_result =
          ((((val & 0x80000000) && 1) ^     /* Bit 31 */
            ((val & 0x20000000) && 1) ^     /* Bit 29 */
            ((val & 0x40000) && 1)    ^     /* Bit 18 */
            ((val & 0x20) && 1)) == 1);     /* Bit  5 */

    output_word = xor_result << 16;

    for(j=0; j<32; j++) {
	if (j==29)
	    continue;
	if (val & (0x1<<j)) 	
	    output_word ^= (0x1 << swap_map[j]);
    }
    return output_word; 
}


void slidas_initialize()
{
    WORD w;
    int i;

    for (i=0; i<65536; ++i) {
	w.sval[0] = i;
	w.sval[1] = 0;
	ltab[i] = slidas_random_next(w.ival);

	w.sval[0] = 0;
	w.sval[1] = i;
	htab[i] = slidas_random_next(w.ival);
    }
}

void slidas_reset()
{
  reg_word.ival = SEED;
}

unsigned int slidas_random()
{
  return reg_word.ival = htab[reg_word.sval[1]] ^ ltab[reg_word.sval[0]];
}

int receive(unsigned int FILAR_CHANNEL, int FRAMELEN, unsigned int *dataptr_solar)
{
  // use static array data[] for filar receive buffer:
  unsigned data[10000];

  // other variables
  int filsize = 0;
  int filerr(0);
  int errcnt = 0;
  while (!filsize) {
    filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
  }
  
  if(filerr!=0) {
    if(false) {
      cerr << "FILAR GROSS ERROR: received " << filsize
	   << " error=" << hex << filerr << dec << endl;
    }
    return -1;
  }
  assert(filsize==FRAMELEN);

  //  return 0; // FIXME

  for(int z=0; z<filsize; z++) {
    if(data[z] != dataptr_solar[z]) {
      errcnt+=1;
      if(false) {
	std::cerr << "FILAR ERROR: sent [" << hex << dataptr_solar[z] 
		  << "], received [" << data[z] << "]" << dec
		  << std::endl;
      }
    }
  }
  return errcnt;
}

int main(unsigned long int argc, char* argv[]) {
  unsigned int pattern(5), mode(0), length(4);
  unsigned int NFRAMES(3000000);
  unsigned int NUMLOOPS(1);
  unsigned int FRAMELEN(254);
  if(argc==1) {
    cout << "USAGE: " << argv[0] << " NFRAMES NUMLOOPS" << endl;
    exit(0);
  }
  if(argc>=2) NFRAMES = str_to_uint(argv[1]);
  if(argc>=3) NUMLOOPS = str_to_uint(argv[2]);
  std::cerr << "pattern = " << pattern << " | mode = " << mode 
	    << " | length = " << length << std::endl;
  std::cerr << "NFRAMES=" << NFRAMES << "; NUMLOOPS="
	    << NUMLOOPS << std::endl;

  srand( time(NULL) ); // random numbers to choose readout order
  reg_word.ival = SEED;
  slidas_initialize();
  
  unsigned int *dataptr_solar = new unsigned int[10000];

  cerr << "SOLAR: before solar_setup" << endl;
  solar_setup(0);
  solar_slidas_off(0);
  cerr << "SOLAR[" << "]: after solar_setup" << endl;
  //  dataptr_solar = solar_getbuffer(0);
  //  cerr << "SOLAR: retrieved solar buffer!"  << endl;

  // Set up two FILAR channels
  cerr << "FILAR: filar_setup with channel mask: " 
       << (FILAR_CHANNEL_MASK_BOTH) << endl;
  int setup_status = filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK_BOTH);
  if(setup_status==-1) {
    cerr << "ERROR: failed to bring up FILAR" << endl;
    exit(0);
  }
  // arm a few REQ FIFOs before going to event loop
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_1, FILAR_ARM);
  int req_armed = FILAR_ARM;

  // enter slidas mode
  solar_slidas_config(0,pattern,length,mode);
  solar_slidas_dump(0);
  
  cerr << "Running tests. "
       << "You should see a status report every few seconds..."
       << endl;

  // prepare some dummy data
  unsigned long int iframe = NFRAMES;
  double errcnt_word[] = {0.0,0.0};
  double errcnt_frame[] = {0.0,0.0};

  struct timeval beg_time, end_time;
  struct timeval beg_time1, end_time1;
  gettimeofday(&beg_time, NULL);
  gettimeofday(&beg_time1, NULL);
#ifdef ADD_SLEEP
  const int CHECKFREQ = 2500;
#else
  const int CHECKFREQ = 250000;
#endif
  for(int iloop=0; iloop<NUMLOOPS; iloop++){
    iframe = NFRAMES;
    while(iframe-- > 0) {
      // periodic status report
      if(iframe%CHECKFREQ==0) {
	gettimeofday(&end_time1, NULL);
	const double nframes = (1.0*iloop)*NFRAMES + (NFRAMES-iframe);
	const double nwords = nframes*(1.0*FRAMELEN);
	const double nbits = nwords*32.0;
	const double dtime = (end_time1.tv_sec - beg_time1.tv_sec) + (end_time1.tv_usec - beg_time1.tv_usec)/1000000.0;
	const double bps = 32.0*FRAMELEN*CHECKFREQ/dtime;
	const double mbps = bps/(1024.0*1024.0*8.0); 
	cout << iloop << "|" << iframe/CHECKFREQ
	     << " | MBs = " << fixed << setprecision(2) << mbps
	     << " | N = " << scientific << setprecision(6) << nbits
	     << fixed << setprecision(0)
	     << " | CH0 = ["<< errcnt_frame[0]
	     << " , " << errcnt_word[0] << "]"
	     << " | CH1 = ["<< errcnt_frame[1]
	     << " , " << errcnt_word[1] << "]"
	     << endl;
	cout << flush;
	beg_time1 = end_time1;
      }
      // pre-fill the SLIDAS check buffer
      if(true) { // FIXME
	for(int z=0; z<FRAMELEN; z++) {
	  dataptr_solar[z] = slidas_random();
	  //dataptr_solar[z] = slidas_random_old(0);
	}
      }
      // receiving each channel
      const bool choose1 = ((rand()%2)==0);
      if(choose1) {
	for(int ir=0;ir<2;ir++) {
	  const int status = receive(ir+1,FRAMELEN,dataptr_solar);
	  if(status == -1) errcnt_frame[ir]+=1.0; else errcnt_word[ir]+=status;
	}
      } else {
	for(int ir=1;ir>=0;ir--) {
	  const int status = receive(ir+1,FRAMELEN,dataptr_solar);
	  if(status == -1) errcnt_frame[ir]+=1.0; else errcnt_word[ir]+=status;
	}
      }
      // re-arm receive FIFOs
      req_armed--;
      if(req_armed==0) {
	const bool choose2 = ((rand()%2)==0);
	if(choose2) {
	  for(int ir=0;ir<2;ir++) {
	    filar_setreq(FILAR_SLOT, ir+1, FILAR_ARM);
#ifdef ADD_SLEEP
	    if(ir==0) usleep(1000);
#endif
	  }
	} else {
	  for(int ir=1;ir>=0;ir--) {
	    filar_setreq(FILAR_SLOT, ir+1, FILAR_ARM);
#ifdef ADD_SLEEP
	    if(ir==1) usleep(500);
#endif
	  }
	}
	req_armed = FILAR_ARM;
      }
    } // NFRAMES loop
  } // NUMLOOPS loop

  gettimeofday(&end_time, NULL);
  cout << "=========================================" << endl;
  cout << "FRAMELEN = " << FRAMELEN << endl;
  cout << "CHANNEL 0: frame errors = " << errcnt_frame[0] 
       << " , words errors = " << errcnt_word[0] << endl;
  cout << "CHANNEL 1: frame errors = " << errcnt_frame[1] 
       << " , words errors = " << errcnt_word[1] << endl;
  const double nframes = (1.0*NFRAMES)*(1.0*NUMLOOPS);
  const double nwords = nframes*(1.0*FRAMELEN);
  const double nbits = nwords*32.0;
  cout << "TOTAL: frames = " << scientific << nframes
       << ", words = " << nwords
       << ", bits = " << nbits << endl;
  //  const double dtime = end_time - beg_time;
  const double dtime = (end_time.tv_sec - beg_time.tv_sec) + (end_time.tv_usec - beg_time.tv_usec)/1000000.0;

  cout << "TOTAL TIME (sec) = " << dtime << endl;
  if(dtime>0) {
    const double bps = 1.0*nbits/dtime;
    cout << "Bit/sec = " << bps << endl;
    cout << "Words/sec = " << nwords/dtime << endl;
    cout << "MB/sec = " << bps/(1024.0*1024.0*8.0) << endl;
    const double tproj = 1e15/bps;
    cout << "Time to test 1e15 bits: " << tproj << " sec, " 
	 << fixed << tproj/(60.0*60.0*24) << " days" << endl;
  }
  cout << "=========================================" << endl;
}


