//
// system headers
//
#include <stdio.h>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <sched.h>
#include <vector>
#include <cassert>

#include "NodeUtils.hh"

//
// s32pci64 headers
//
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

int HEADER[18] = {0xdd1234dd,
		0x338,
		0x9,
		0x4000000,
		0x210002,
		0x2,
		0x0,
		-1,
		0xee1234ee,
		0x9,
		0x3010000,
		0x200000,
		0x0,
		-1,
		0xbc1d,
		0x0,
		0x18000};

int TAILER[6] = {0x0,
		 0x40000,
		 0x2,
		 0x320,
		 0x1,
		 -1};

using namespace std;

#define FILAR_NUMBER_0 0

#define rdtscl(low) \
     __asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

unsigned int str_to_uint(string v) {
  unsigned int result;
  istringstream buf(v);
  buf >> result;
  return result;
}
int chan_to_mask(int ch) {
  return 1 << (ch-1);
}

int main( int argc, char* argv[]) 
{
  int PRERUN=200;
  int flow=true;
  int dump=2;
  bool validate = false; // validate bit-walking pattern?

  int filar_err;

  unsigned int FILAR_CHANNEL(0x1);
  unsigned int FILAR_CHANNEL_MASK = chan_to_mask(FILAR_CHANNEL);

  if(argc>=2) {
    flow=false;
  }

  cerr << "CHANNEL AND MASK: " << FILAR_CHANNEL 
       << " and " << FILAR_CHANNEL_MASK << endl;
  assert(FILAR_CHANNEL>0 && FILAR_CHANNEL<4);

  unsigned data[10000];
  unsigned  *dataptr_filar;
  unsigned  *dataptr_solar;
  dataptr_filar = &(data[0]);
  
  cerr << "before filar_setup"  << endl;   
  filar_ak0_init(FILAR_NUMBER_0);
  int card_status = filar_ak1_cardreset(FILAR_NUMBER_0);
  if(card_status==-1) {
    cerr << "ERROR: failed to bring up FILAR *card*" << endl;
    exit(0);
  }
  int link_status = filar_ak2_linkreset_forever(FILAR_NUMBER_0, FILAR_CHANNEL_MASK); 
  if(link_status==-1) {
    cerr << "ERROR: failed to bring up FILAR" << endl;
    exit(0);
  }
  else {
    cerr << "after filar_setup"  << endl;
  }
  cout << endl;
  
  int nread = 0;
  const int FILAR_ARM = 15;

  // ARM REQ FIFOs before going to event loop
  filar_setreq(FILAR_NUMBER_0, FILAR_CHANNEL, FILAR_ARM);

  cerr << "Link is up..." << endl;
  if(flow) {
    cerr << "Press ENTER to enable FTK flow control... ";
  } else  {
    cerr << "Press ENTER to disable FTK flow control... ";
  }
  cin.ignore();
  if(flow) {
    filar_flowctrlenable(FILAR_NUMBER_0, FILAR_CHANNEL);
    cerr << "FTK flow control enabled!" << endl;
  }
  else {
    filar_flowctrldisable(FILAR_NUMBER_0, FILAR_CHANNEL);
    cerr << "FTK flow control disabled!" << endl;
  }
  cerr << "Waiting for 200 events..." << endl;

  //  event loop
  long int ievent = 0;
  long iword = 0; // within a bit-walking frame
  long ipat = 0;
  while(PRERUN-->1) {
    // wait for data on the filar
    filar_err=0;
    // re-arm FILAR FIFOs
    if (nread == FILAR_ARM-1){
      filar_setreq(FILAR_NUMBER_0, FILAR_CHANNEL, FILAR_ARM);
    }
    int wlcnt=0;
    int fsize=0;
    while (!fsize) {
      fsize=filar_receive_channel_ptr( FILAR_NUMBER_0, &filar_err, FILAR_CHANNEL, (unsigned*)(&data));
      wlcnt++;
    }
    if(dump) {
      cout << "Event # " << ievent 
	   << " with " << fsize << " words" << endl;
    }
    if(dump>=2) {
      for( int j=0; j<fsize; j++ ) {
	cout << "word " << dec << j << " : 0x" 
	     << hex <<  *(dataptr_filar+j) 
	     << dec << endl;
      }
    }
    if(dump) {
      cout << endl;
      cout << flush;
    }
    if(validate) {
      const int NTIMES=25;
      const int NPAT=32;
      const int hst = 17;
      const int tst = NTIMES*NPAT+17+1;
      assert(fsize==tst+6);
      for( int j=0; j<fsize; j++ ) {
	int word = *(dataptr_filar+j);
	bool passed_HEAD=false,passed_DATA=false,passed_TAIL=false;
	if(iword<=hst) {
	  if(word==HEADER[iword] || HEADER[iword]==-1) passed_HEAD=true;
	}
	if(iword>hst && iword<tst) {
	  int ipat = iword-18;
	  if((1<<ipat) == word) passed_DATA = true;
	}
	if(iword>=tst) {
	  if(word==TAILER[iword-tst] || TAILER[iword-tst]==-1) passed_TAIL=true;
	}
	iword++;
	bool passed = passed_HEAD || passed_DATA || passed_TAIL;
	if(!passed) cerr << "ERROR: failed word: " << word << endl;
      }
    }
    nread++;
    if (nread == FILAR_ARM) {
      nread = 0;
    }
    ievent++;
  } // while true

  cerr
    << "Stopping event readout; this should assert FTK flow control" 
    << endl;
  cerr << "Will quit automatically after 20 seconds..." << endl;

  sleep(30);
  filar_flowctrldisable(FILAR_NUMBER_0, FILAR_CHANNEL);
  cerr << "DONE! Disabling FTK flow control and quitting" << endl;
  
}
