//
// system headers
//
#include <stdio.h>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <sched.h>
#include <vector>
#include <cassert>

#include "NodeUtils.hh"

//
// s32pci64 headers
//
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

using namespace std;

#define FILAR_NUMBER_0 0

#define rdtscl(low) \
     __asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

unsigned int str_to_uint(string v) {
  unsigned int result;
  istringstream buf(v);
  buf >> result;
  return result;
}
int chan_to_mask(int ch) {
  return 1 << (ch-1);
}
int main( int argc, char* argv[]) 
{  
  unsigned int FILAR_CHANNEL(0x2);
  unsigned int FILAR_CHANNEL_MASK(0x2);

  if(argc==2) {
    FILAR_CHANNEL = str_to_uint(argv[1]);
    FILAR_CHANNEL_MASK = chan_to_mask(FILAR_CHANNEL);
    cerr << "NEW CHANNEL AND MASK: " << FILAR_CHANNEL 
	 << " and " << FILAR_CHANNEL_MASK << endl;
  }
  assert(FILAR_CHANNEL>0 && FILAR_CHANNEL<4);

  unsigned data[10000];
  unsigned  *dataptr_filar;
  unsigned  *dataptr_solar;
  dataptr_filar = &(data[0]);

  cerr << "before filar_setup"  << endl;   
  int setup_status = filar_setup_hola(FILAR_NUMBER_0, FILAR_CHANNEL_MASK); 
  if(setup_status==-1) {
    cerr << "ERROR: failed to bring up FILAR" << endl;
    exit(0);
  }
  else {
    cerr << "after filar_setup"  << endl;
  }

  int reqfifobufcount = 0;
  int reqfifobufused = 1;

  // ARM REQ FIFOs before going to event loop
  filar_setreq(FILAR_NUMBER_0, FILAR_CHANNEL, 4);

  //  event loop
  int nevent=1;
  int filar_err;
  int count=0, dumpcount=0;
  while( 1 )
    {
      // wait for data on the filar
      filar_err=0;
      int fsize=0;
      // Write REQ FIFO for Cluster channel
      if (reqfifobufcount == reqfifobufused-1){
	filar_setreq(FILAR_NUMBER_0, FILAR_CHANNEL, reqfifobufused);
      }
      int wlcnt=0;
      while (!fsize)
 	{
	  fsize=filar_receive_channel_ptr( FILAR_NUMBER_0, &filar_err, FILAR_CHANNEL, (unsigned*)(&data));
	  wlcnt++;
 	}
      printf("Made it out alive\n");
      for( int j=0; j<fsize; j++ )
	cout << "word " << dec << j << " : 0x" << hex <<  *(dataptr_filar+j) << endl;
      count++;
      cout << endl;
      cout << flush;
      reqfifobufcount++;
      if (reqfifobufcount == reqfifobufused) {
	reqfifobufcount = 0;
      }
    } // while 1
}
