#include <unistd.h>
#include <sched.h>
#include <pthread.h>
#include <iomanip.h>
#include <signal.h>
#include <linux/types.h>
#include <iostream>
#include <vector>
#include <string.h>
#include <getopt.h>

#include "NodeUtils.hh"

extern "C" {
#include "s32pci64-filar.h"
}


void usage()
{
  cout << "Script to test the channel link on the FILAR\n";
  cout << "Usage: " << endl;
  cout << "--filar0 channels --filar0 channels --filar0 channels\n";
  cout << "Example: filarreset.exe --filar0 201 --filar1 0" << endl;
  cout << "will bring up channels 0,1,2 on filar 0 and channel 0 on filar 1\n";

  
}


int main(int argc, char **argv)
{

  //  int filar0;
  std::string filar0;
  std::string filar1;
  std::string filar2;
  __u32 mask0 = 0;
  __u32 mask1 = 0; 
  __u32 mask2 = 0;
  char o;  
  int option_index = 0;
  static struct option long_options[] =
    {
      {"filar0", required_argument, 0, 'a'},
      {"filar1", required_argument, 0, 'b'},
      {"filar2", required_argument, 0, 'c'},
      {"help", no_argument, 0, 'h'}
    };

  while ((o=getopt_long(argc,argv,"a:b:c:h"
			,long_options, &option_index)) != EOF) {


    switch (o) {
    case 'a':
      filar0 = std::string(optarg);
      break;
    case 'b':
      filar1 = std::string(optarg);
      break;
    case 'c':
      filar2 = std::string(optarg);
      break;
    default :
    case 'h':
      // get rid of "for :"
      usage();
      return 1;
      // break;
    }
  }
  
  int pos = 0;

  if (filar0.length() > 0)
    cerr << "On FILAR 0 bring up channels ";
  
  for (int i = 0; i < filar0.length() ; i++){
    pos = (int)filar0[i] - 48; // account for ascii to number conversion (48)
    cerr << pos << ", ";
    if (pos > 3 ) {
      cerr << "Channel above range (3)"<<endl;
      exit(1);
    }
    mask0 = mask0 | (1<< pos);
  }
  cerr << endl;

  if (filar1.length() > 0)
    cerr << "On FILAR 1 bring up channels ";
  
  for (int i = 0; i < filar1.length() ; i++){
    pos = (int)filar1[i] - 48; // account for ascii to number conversion (48)
    cerr << pos << ", ";
    if (pos > 3 ) {
      cerr << "Channel above range (3)"<<endl;
      exit(1);
    }
    mask1 = mask1 | (1<< pos);
  }
  cerr << endl;

  if (filar2.length() > 0)
    cerr << "On FILAR 0 bring up channels ";
  
  for (int i = 0; i < filar2.length() ; i++){
    pos = (int)filar2[i] - 48; // account for ascii to number conversion (48)
    cerr << pos << ", ";
    if (pos > 3 ) {
      cerr << "Channel above range (3)"<<endl;
      exit(1);
    }
    mask2 = mask2 | (1<< pos);
  }
  cerr << endl;
  
  
  if (filar0.length() > 0){
    cerr << "setting up filar0 ... "<< endl;
    filar_setup_hola(0, ((mask0)&0xf) );
  }

  if (filar1.length() > 0){
    cerr << "setting up filar1 ... "<< endl;
    filar_setup_hola(1, ( (mask1)&0xf) );
  }

  if (filar2.length() > 0){
    cerr << "setting up filar2 ... "<< endl;
    filar_setup_hola(2, ( (mask2)&0xf) );
  }
  


}
