//
// compile with : "g++ -I l2pulsar_slink/s32pci64-filar/s32pci64-filar/ filarreset.cc -o filarreset.exe -L ./lib/ -ll2slink"
//

#include <unistd.h>
#include <sched.h>
#include <pthread.h>
#include <iomanip.h>
#include <signal.h>
#include <linux/types.h>
#include <iostream>
#include <vector>
#include <string.h>

#include "include/NodeUtils.hh"


extern "C" {
#include "s32pci64-filar.h"
}



//
// s32pci64 defines
//



// two buses
#define FILAR_NUMBER_0   0x00   // right from SOLAR
#define FILAR_NUMBER_1   0x01  // left from SOLAR
#define FILAR_NUMBER_2   0x02

// old system

//#define FILAR_CHANNEL_CLUSTER   0x01
#define FILAR_CHANNEL_CLUSTER   0x01
#define FILAR_CHANNEL_MUON   0x02
#define FILAR_CHANNEL_RECES   0x03
#define FILAR_CHANNEL_SVT   0x04


//#define CLUSTER_MASK   0x01
#define CLUSTER_MASK   0x10
#define MUON_MASK   0x02
#define RECES_MASK   0x04
#define SVT_MASK   0x08

// l2cal

#define FILAR_CHANNEL_NEWCLUSTER_1   0x04
#define FILAR_CHANNEL_NEWCLUSTER_2   0x03
#define FILAR_CHANNEL_NEWCLUSTER_3   0x02


#define NEWCLUSTER_MASK_1 0x80
#define NEWCLUSTER_MASK_2 0x40
#define NEWCLUSTER_MASK_3 0x20
//xft

#define FILAR_CHANNEL_XFT_1 0x04
#define FILAR_CHANNEL_XFT_2 0x03
#define FILAR_CHANNEL_XFT_3 0x02

#define XFT_MASK_1 0x800
#define XFT_MASK_2 0x400
#define XFT_MASK_3 0x200



/////////////////////////////////////////////////

#define FILAR_BUFFER_SIZE     2048
#define SOLAR_BUFFER_SIZE     2048


#define ALL_CHANNEL_MASK 0xFF



using namespace std;

vector<__u32> filarDataBig( FILAR_BUFFER_SIZE );
vector<__u32> filarDataLittle( FILAR_BUFFER_SIZE );
__u32* filarptr = (__u32*)&(filarDataBig[0]);

__u32 ch1Time_begin = 0;
__u32 ch1Time_end = 0;
__u32 ch2Time_begin = 0;
__u32 ch2Time_end = 0;
__u32 ch3Time_begin = 0;
__u32 ch3Time_end = 0;
__u32 ReqTime_begin = 0;
__u32 ReqTime_end = 0;


void catch_int(int sig_num)
{
  // re-set handler to catch_int, for the "next time" 
  signal(SIGINT, catch_int);

  cerr << endl;
  cerr << endl;
  cerr << "dying gracefully ..." << endl;
  cerr << endl;

  filar_unmap(0);
  filar_exit(0);

  cerr << "finished." << endl;
  exit(0);
}


int main()
{
  __u32 channel_mask;

  int size1, size2, size3, filar_err, count, i;
  int ret;
  vector<__u32> newclusterData1( FILAR_BUFFER_SIZE );
  vector<__u32> newclusterData2( FILAR_BUFFER_SIZE );
  vector<__u32> newclusterData3( FILAR_BUFFER_SIZE );
  __u32* clusterptr_new_1 = (__u32*)&(newclusterData1[0]); 
  __u32* clusterptr_new_2 = (__u32*)&(newclusterData2[0]); 
  __u32* clusterptr_new_3 = (__u32*)&(newclusterData3[0]); 

  //  channel_mask = NEWCLUSTER_MASK_1 | NEWCLUSTER_MASK_2 | NEWCLUSTER_MASK_3;

  channel_mask = NEWCLUSTER_MASK_1 | NEWCLUSTER_MASK_2 | NEWCLUSTER_MASK_3 | CLUSTER_MASK | SVT_MASK | RECES_MASK | MUON_MASK | XFT_MASK_1 | XFT_MASK_2 | XFT_MASK_3;

  //  channel_mask = channel_mask | CLUSTER_MASK | MUON_MASK | RECES_MASK | SVT_MASK;
  //  channel_mask = MUON_MASK | CLUSTER_MASK;

  //  channel_mask = NEWCLUSTER_MASK_1 | NEWCLUSTER_MASK_2;

  // 
  // s32pci64 stuff
  // ----------------------------------
  cerr << "setting up filar ... "<< endl;	
  //  signal(SIGINT, catch_int);
  // ----------------------------------  
  


  cerr << "setting up filar0 ... "<< endl;
  filar_setup_hola(0, ((channel_mask)&0xf) );

  cerr << "setting up filar1 ... "<< endl;
  filar_setup_hola(1, ( (channel_mask>>4)&0xf) );

  cerr << "setting up filar2 ... "<< endl;
  filar_setup_hola(2, ( (channel_mask>>8)&0xf) );



  // Write REQ FIFO for Cluster channel
  count = 0 ;

  while (count < 15) {


    cerr << "count= " << count <<endl;
    // ARM REQ FIFOs before going to event loop
    //
    filar_setreq(FILAR_NUMBER_0, FILAR_CHANNEL_MUON, 1);
    filar_setreq(FILAR_NUMBER_0, FILAR_CHANNEL_RECES, 1);
    filar_setreq(FILAR_NUMBER_0, FILAR_CHANNEL_SVT, 1);
    //    filar_setreq(FILAR_NUMBER_0, FILAR_CHANNEL_CLUSTER, 1);
    filar_setreq(FILAR_NUMBER_1, FILAR_CHANNEL_CLUSTER, 1);

    cerr << "FILAR_NUMBER_1, FILAR_CHANNEL_NEWCLUSTER_1" << endl;
    filar_setreq(FILAR_NUMBER_1, FILAR_CHANNEL_NEWCLUSTER_1, 1);
    cerr << "FILAR_NUMBER_1, FILAR_CHANNEL_NEWCLUSTER_2" << endl;
    filar_setreq(FILAR_NUMBER_1, FILAR_CHANNEL_NEWCLUSTER_2, 1);
    cerr << "FILAR_NUMBER_1, FILAR_CHANNEL_NEWCLUSTER_3" << endl;
    filar_setreq(FILAR_NUMBER_1, FILAR_CHANNEL_NEWCLUSTER_3, 1);

    cerr << "FILAR_NUMBER_2, FILAR_CHANNEL_XFT_1" << endl;
    filar_setreq(FILAR_NUMBER_2, FILAR_CHANNEL_XFT_1, 1);
    cerr << "FILAR_NUMBER_2, FILAR_CHANNEL_XFT_2" << endl;
    filar_setreq(FILAR_NUMBER_2, FILAR_CHANNEL_XFT_2, 1);
    cerr << "FILAR_NUMBER_2, FILAR_CHANNEL_XFT_3" << endl;
    filar_setreq(FILAR_NUMBER_2, FILAR_CHANNEL_XFT_3, 1);

    //  rdtscl(ReqTime_end);


    size1 = 0;
    while( size1 == 0) {
      rdtscl(ch1Time_begin);
      size1=filar_receive_channel_ptr( FILAR_NUMBER_1, &filar_err, FILAR_CHANNEL_CLUSTER, clusterptr_new_1);
      rdtscl(ch1Time_end);
    }
    cerr << "size1 = " << size1 <<" ch1 Time (100ns)" << dec <<tstamp_to_100ns(ch1Time_begin,ch1Time_end)<< endl;
    //    sleep(1);

    size2 = 0;
    while( size2 == 0) {
      rdtscl(ch2Time_begin);
      size2=filar_receive_channel_ptr( FILAR_NUMBER_0, &filar_err, FILAR_CHANNEL_MUON, clusterptr_new_2);
      rdtscl(ch2Time_end);
    }

    cerr << "size2 = " << size2 <<" ch2 Time (100ns)" << dec <<tstamp_to_100ns(ch2Time_begin,ch2Time_end)<< endl;

    /*
    size3 = 0;
    while( size3 == 0) {
      rdtscl(ch3Time_begin);
      size3=filar_receive_channel_ptr( FILAR_NUMBER_1, &filar_err, FILAR_CHANNEL_NEWCLUSTER_3, clusterptr_new_3);
      rdtscl(ch3Time_end);
    }



    cerr << "size3 = " << size3 <<" ch3 Time (100ns)" << dec <<tstamp_to_100ns(ch3Time_begin,ch3Time_end)<< endl;
    */ 
    count++;

    // print out data we received


    for (i = 0 ; i< size1; i++){
      cerr<< "Data1[" << i << "]" << hex << *(clusterptr_new_1+i) << dec << endl;
    }


    for (i = 0 ; i< size2; i++){
      cerr<< "Data2[" << i << "]" << hex << *(clusterptr_new_2+i) << dec<< endl;
    }

    /*

    for (i = 0 ; i< size3; i++){
      cerr<< "Data3[" << i << "]" << hex << *(clusterptr_new_3+i) << dec << endl;
    }
    */

  }
  
}
