#include <stdio.h>
#include <cstdlib>
#include <cassert>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include <sys/time.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

//#define ADD_SLEEP

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 
#define FILAR_CHANNEL_0 0x1
#define FILAR_CHANNEL_1 0x2
#define FILAR_CHANNEL_MASK_0 0x1
#define FILAR_CHANNEL_MASK_1 0x2

#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

unsigned int str_to_uint(string v) {
  unsigned int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

int main(unsigned long int argc, char* argv[]) {
  int urlchannel = 2;
  if(argc<2) {
    cout << "USAGE: " << argv[0] << " CHANNEL_MASK" << endl;
    exit(0);
  }
  if(argc>=2) urlchannel = str_to_uint(argv[1]);
  bool breakout=false;
  if(argc>2) breakout=true;

  assert(urlchannel==FILAR_CHANNEL_MASK_0 || urlchannel==FILAR_CHANNEL_MASK_1 || urlchannel==FILAR_CHANNEL_MASK_BOTH);

  unsigned int FILAR_CHANNEL = urlchannel;

  solar_setup(0);
  solar_slidas_off(0);
  // Set up both FILAR channels
  cerr << "FILAR: filar_setup with channel mask: " 
       << (FILAR_CHANNEL_MASK_BOTH) << endl;
  int setup_status = filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK_BOTH);
  if(setup_status==-1) {
    cerr << "ERROR: failed to bring up FILAR" << endl;
    exit(0);
  }

  cout << "====================================================" << endl;
  cout << "FTK_XOFF_ENA LED should start blinking now..." << endl;
  cout << "Press CTRL-C when done!" << endl;
  cout << "====================================================" << endl;

  // just run in an infinite loop to make the FTK_XOFF_ENA LED blink
  bool do_enable=false;
  int i=0;
  while(true) {
    if(do_enable) { // currently disabled
      filar_flowctrlenable(0,FILAR_CHANNEL);
      do_enable=false;
    }
    else { // currently enabled
      filar_flowctrldisable(0,FILAR_CHANNEL);
      do_enable=true;
    }
    usleep(1.0*10000);
    if(breakout && i==1)  break;
    i++;
  }
  return(0);
}
