#include <stdio.h>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

#define BOF_WORD BOF0AAA0
#define EOF_WORD EOF0AAA0

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 

// first FILAR channel:
#if 0
#define FILAR_CHANNEL_MASK 0x1
#define FILAR_CHANNEL_0 0x1
#else
// OR: second FILAR channel:
#define FILAR_CHANNEL_MASK 0x2
#define FILAR_CHANNEL_0 0x2
#endif
#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

#define rdtscl(low) \
__asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

int timing() {
    static __u32 told;
    __u32 tnew;
    int result;
    rdtscl(tnew);
    result = tstamp_to_us(told,tnew);
    told = tnew;
    return result;
}

int str_to_int(string v) {
  int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

void receive(unsigned int FILAR_CHANNEL) {
    // use static array data[] for filar receive buffer:
    unsigned data[100000];
    // other variables
    int filsize = 0;
    int filerr(0);

    timing();
    while (!filsize) {
      filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
    }
    
    cerr << "FILAR[" << timing() << "]: received " << filsize 
	 << " words; error=" << hex << filerr << dec << endl;
    if(filerr!=0) {
      cerr << "ERROR: major frame error" << endl;
      exit(0);
    }
    for(int z=0; z<filsize; z++) {
      if(z>1 && z<filsize-2) continue; // for brevity
        cerr << "FILAR: data[" << z << "] = " 
            << hex << data[z] << dec << endl;
    }
}

int main( int argc, char* argv[]) {
  // length of frame that is sent through SOLAR
  int FRAMELEN(3);
  int NFRAMES(2);
  if(argc<2) {
    cout << "USAGE: " << argv[0] << " FRAMELEN NFRAMES" << endl;
    //exit(0);
  }
  if(argc>=2) FRAMELEN = str_to_int(argv[1]);
  if(argc>=3) NFRAMES = str_to_int(argv[2]);
  cout <<"FRAMELEN=" << FRAMELEN << "; NFRAMES=" << NFRAMES << endl;
  
  unsigned int *dataptr_solar;

  cerr << "SOLAR: before solar_setup" << endl;
  timing();
  solar_setup(0);
  cerr << "SOLAR[" << timing() << "]: after solar_setup" << endl;
  dataptr_solar = solar_getbuffer(0);
  cerr << "SOLAR: retrieved solar buffer!"  << endl;
 
  // We have a faulty SOLAR that doesn't always come up
  // The way to detect failure is by looking at FILAR LDOWN
  // If detected, we re-reset SOLAR and try again.
  int linkup=0;
  for(int itry=0; itry<3; itry++) {
    cerr << "FILAR: before filar_setup (attempt " << itry << ")" << endl;   
    timing();
    if( filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK) == -1) {
      solar_setup(0);
      cerr << "SOLAR[" << timing() << "]: after solar_setup" << endl;
      dataptr_solar = solar_getbuffer(0);
    }
    else {
      cerr << "FILAR[" << timing() << "] after filar_setup" << endl;
      linkup=1;
      break;
    }
  }
  if(linkup==0) {
    cerr << "ERROR: unable to bring up FILAR" << endl;
    exit(0);
  }
  // arm a few REQ FIFOs before going to event loop
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
  int req_armed = FILAR_ARM;
  
  cerr << "STATS: done initialization" << endl;

  // prepare some dummy data
  int sent=0;
  for(int il=0;il<FRAMELEN;il++) {*(dataptr_solar+il)=0xaaaaaaaa+il;}

  cerr << "SOLAR: sending " << FRAMELEN << " words" << endl;
  timing();
  solar_send(FRAMELEN,0); sent++;
  cerr << "SOLAR[" << timing() << "]: sent " << FRAMELEN << " words" << endl;
  cerr << "---> sent count = " << sent << endl;
  
  // now try to trigger XOFF from FILAR end
  cerr << "SOLAR: sending data train of " << NFRAMES-1 << " frames" << endl;
  for(int il=0;il<NFRAMES-1;il++) {
    *(dataptr_solar+FRAMELEN-1) = 0xabcdaaaa+il;
    solar_send(FRAMELEN,0); sent++;
    cerr << "---> sent count = " << sent << endl;
  }

  // Continuous loop to control data reception
  cerr << "FYI, sent a total of " << sent << " frames ("
       << sent*FRAMELEN << ") words" << endl;
  while(true) {
    int nreceive;
    cerr << "Enter number of buffers to receive: "; 
    cin >> nreceive;
    for(int ir=0;ir<nreceive;ir++) {
      receive(FILAR_CHANNEL_0);
      req_armed--;
      // time to re-arm the REQ FIFO
      if(req_armed==0) {
	cerr << "INFO: Arming " << FILAR_ARM << " REQ FIFO buffers..." << endl;
	filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
	sleep(0.1);
	req_armed = FILAR_ARM;
	cerr << "Armed!" << endl;
      }
    }
  }
}
