#include <stdio.h>
#include <cstdlib>
#include <cassert>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include <sys/time.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

//#define ADD_SLEEP

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 
#define FILAR_CHANNEL_0 0x1
#define FILAR_CHANNEL_1 0x2
#define FILAR_CHANNEL_MASK_0 0x1
#define FILAR_CHANNEL_MASK_1 0x2

#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

#define rdtscl(low) \
__asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

int timing() {
    static __u32 told;
    __u32 tnew;
    int result;
    rdtscl(tnew);
    result = tstamp_to_us(told,tnew);
    told = tnew;
    return result;
}

unsigned int str_to_uint(string v) {
  unsigned int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

int slidas_random_old(int reset)
{
  static unsigned reg_word = 0x524948e0; /* Random number seed */
  unsigned output_word;
  int i, j, xor_result;
  static int swap_map[32] = { 19,  0,  9, 20,  5, 31, 24, 22, 
			      1,  7, 27,  8, 17, 29, 23, 11, 
			      18, 21, 25,  2,  6, 13,  3, 10, 
			      30, 28, 14, 12,  4,  0, 26, 15};
  if (reset) reg_word = 0x524948e0;				   
  xor_result = 
    ((((reg_word & 0x80000000) && 1) ^     /* Bit 31 */
      ((reg_word & 0x20000000) && 1) ^     /* Bit 29 */
      ((reg_word & 0x40000) && 1)    ^     /* Bit 18 */
      ((reg_word & 0x20) && 1)) == 1);     /* Bit  5 */
  output_word = 0;
  for(j=0; j<32; j++) {
    if (j != 29) {
      if (j == 16)		 	    
	output_word ^= (xor_result << 16); 
      if (reg_word & (0x1 << j)) 	
	output_word ^= (0x1 << swap_map[j]);
    }
  }
  reg_word = output_word;
  return output_word; 
}

#define SEED 0x524948e0 /* Random number seed */

typedef union {
    unsigned int ival;
    unsigned short int sval[2];
} WORD;
    
static WORD reg_word;

unsigned int htab[65536], ltab[65536];

unsigned int slidas_random_next(unsigned int val)
{
    unsigned output_word, xor_result;
    int j;
    static int swap_map[32] = { 19,  0,  9, 20,  5, 31, 24, 22, 
				 1,  7, 27,  8, 17, 29, 23, 11, 
				18, 21, 25,  2,  6, 13,  3, 10, 
				30, 28, 14, 12,  4,  0, 26, 15 };
    

    xor_result =
          ((((val & 0x80000000) && 1) ^     /* Bit 31 */
            ((val & 0x20000000) && 1) ^     /* Bit 29 */
            ((val & 0x40000) && 1)    ^     /* Bit 18 */
            ((val & 0x20) && 1)) == 1);     /* Bit  5 */

    output_word = xor_result << 16;

    for(j=0; j<32; j++) {
	if (j==29)
	    continue;
	if (val & (0x1<<j)) 	
	    output_word ^= (0x1 << swap_map[j]);
    }
    return output_word; 
}


void slidas_initialize()
{
    WORD w;
    int i;

    for (i=0; i<65536; ++i) {
	w.sval[0] = i;
	w.sval[1] = 0;
	ltab[i] = slidas_random_next(w.ival);

	w.sval[0] = 0;
	w.sval[1] = i;
	htab[i] = slidas_random_next(w.ival);
    }
}

void slidas_reset()
{
  reg_word.ival = SEED;
}

unsigned int slidas_random()
{
  return reg_word.ival = htab[reg_word.sval[1]] ^ ltab[reg_word.sval[0]];
}

int receive(unsigned int FILAR_CHANNEL, int FRAMELEN, unsigned int *dataptr_solar, int print)
{
  // use static array data[] for filar receive buffer:
  unsigned data[10000];

  // other variables
  int filsize = 0;
  int filerr(0);
  int errcnt = 0;
  while (!filsize) {
    filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
  }

  // print received data (only from channel 1)
  	if (print==1 & FILAR_CHANNEL==1)
	{
	  cout <<"data received:"<<endl;
 	  for (int i=0; i<filsize; i++)
 	  {
 		cout<<data[i]<<", ";
	  } 
	  cout<<endl;
	}

  if(filerr!=0) {
    if(false) {
      cerr << "FILAR GROSS ERROR: received " << filsize
	   << " error=" << hex << filerr << dec << endl;
    }
    return -1;
  }
  assert(filsize==FRAMELEN);

  //  return 0; // FIXME

  for(int z=0; z<filsize; z++) {
    if(data[z] != dataptr_solar[z]) {
      errcnt+=1;
      if(false) {
	std::cerr << "FILAR ERROR: sent [" << hex << dataptr_solar[z] 
		  << "], received [" << data[z] << "]" << dec
		  << std::endl;
      }
    }
  }
  return errcnt;
}



int main()
{
// rewrite a new version with more options
	//unsigned long int argc, char* argv[]) {
  unsigned int pattern(8), mode(0), length(4);
  unsigned int NFRAMES(3000000);
  unsigned int NUMLOOPS(1);
  unsigned int FRAMELEN(254);
  int patternop, FRAMELENop;
  int print;
  int dummy(1);

// input parameters:
  cout << "Welcome to the dualrawslidas program. Please set up the parameters and be careful not to exceed the memory limit of the machine."<<endl;
  cout << "Please select the mode in which you would like to operate"<<endl;
  cout << "1: random data"<<endl;
  cout << "2: ROD mode"<<endl;
  cout << "3: ROD mode with 4 lengths at a time (frame length <= 512)" << endl;
  cout << "4: test mode"<<endl;
  cout << "5: Hola test mode"<<endl;
  //you will see an error for each frame because of the missing start word, and the frame length does not matter in this option (length=33 always)
  cin >> patternop;
  cout << "Please set the length of each frame: (16/64/128/256/512/1024)"<<endl;
  cin >> FRAMELENop;
  cout << "Please set the number of frames you would like to send" << endl;
  cin >> NFRAMES;
  cout << "Please set the number of times you want to run the test"<< endl;
  cin >> NUMLOOPS;
  cout << "Do you want to print the sent and received data? (yes=1/no=0)" << endl;
  cin >> print;
// slidas setup  
// pattern setup
  if (patternop==3) pattern=9;
  if (patternop==1)  // random data with control word 
  	{
  		pattern=5;
		srand( time(NULL) ); // random numbers to choose readout order
  	}	
  if (patternop==2) pattern=8; //ROB with 1 block at a time
  if (patternop==4) pattern=0;
	
  if (pattern==9 and length>1) FRAMELEN=FRAMELENop+FRAMELENop/4-2;
  else if (pattern==0) FRAMELEN=32;
  else FRAMELEN=FRAMELENop-2;

//length setup
  if (FRAMELENop==16) length=1;
  else if (FRAMELENop==64) length=2;
  else if (FRAMELENop==128) length=3;
  else if (FRAMELENop==256) length=4;
  else if (FRAMELENop==512) length=5;
  else length=6;
 
  std::cerr << "pattern = " << pattern << " | mode = " << mode 
	    << " | length = " << length << std::endl;
  std::cerr << "NFRAMES=" << NFRAMES << "; NUMLOOPS="
	    << NUMLOOPS << std::endl;
  reg_word.ival = SEED;
  slidas_initialize();
  
  unsigned int *dataptr_solar = new unsigned int[10000];

  cerr << "SOLAR: before solar_setup" << endl;
  solar_setup(0);
  solar_slidas_off(0);
  cerr << "SOLAR[" << "]: after solar_setup" << endl;
  //  dataptr_solar = solar_getbuffer(0);
  //  cerr << "SOLAR: retrieved solar buffer!"  << endl;

  // Set up two FILAR channels
  cerr << "FILAR: filar_setup with channel mask: " 
       << (FILAR_CHANNEL_MASK_BOTH) << endl;
  int setup_status = filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK_BOTH);
  if(setup_status==-1) {
    cerr << "ERROR: failed to bring up FILAR" << endl;
    exit(0);
  }
  // arm a few REQ FIFOs before going to event loop
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_1, FILAR_ARM);
  int req_armed = FILAR_ARM;

  // enter slidas mode
  solar_slidas_config(0,pattern,length,mode);
  solar_slidas_dump(0);
  
  cerr << "Running tests. "
       << "You should see a status report every few seconds..."
       << endl;

  // prepare some dummy data
  unsigned long int iframe = NFRAMES;
  double errcnt_word[] = {0.0,0.0};
  double errcnt_frame[] = {0.0,0.0};

  struct timeval beg_time, end_time;
  struct timeval beg_time1, end_time1;
  gettimeofday(&beg_time, NULL);
  gettimeofday(&beg_time1, NULL);
#ifdef ADD_SLEEP
  const int CHECKFREQ = 2500;
#else
  const int CHECKFREQ = 250000;
#endif
  
  for(int iloop=0; iloop<NUMLOOPS; iloop++)
  {
    iframe = NFRAMES;
    int remainder = NFRAMES%4;
    while(iframe-- > 0) 
    {
      // periodic status report
      if(iframe%CHECKFREQ==0) 
      {
	gettimeofday(&end_time1, NULL);
	const double nframes = (1.0*iloop)*NFRAMES + (NFRAMES-iframe);
	const double nwords = nframes*(1.0*FRAMELEN);
	const double nbits = nwords*32.0;
	const double dtime = (end_time1.tv_sec - beg_time1.tv_sec) + (end_time1.tv_usec - beg_time1.tv_usec)/1000000.0;
	const double bps = 32.0*FRAMELEN*CHECKFREQ/dtime;
	const double mbps = bps/(1024.0*1024.0*8.0); 
	cout << iloop << "|" << iframe/CHECKFREQ
	     << " | MBs = " << fixed << setprecision(2) << mbps
	     << " | N = " << scientific << setprecision(6) << nbits
	     << fixed << setprecision(0)
	     << " | CH0 = ["<< errcnt_frame[0]
	     << " , " << errcnt_word[0] << "]"
	     << " | CH1 = ["<< errcnt_frame[1]
	     << " , " << errcnt_word[1] << "]"
	     << endl;
	cout << flush;
	beg_time1 = end_time1;  
      }

      // pre-fill the SLIDAS check buffer

      if(pattern==8 or pattern==9)
      // data for ROB mode
      {
	dataptr_solar[0]=3994170606;
	dataptr_solar[1]=32;
	dataptr_solar[2]=33685504;	
	dataptr_solar[3]=0;
	dataptr_solar[4]=NFRAMES-iframe+iloop*remainder;
	dataptr_solar[5]=((NFRAMES-iframe+iloop*remainder)*3-2)%4096;
	dataptr_solar[6]=0;
	dataptr_solar[7]=0;
	dataptr_solar[8]=0;
	dataptr_solar[9]=0;
	for(int z=1; z<FRAMELEN-12; z++)
	{
	 dataptr_solar[z+9]=z;
	}
	dataptr_solar[FRAMELEN-3]=1;
	dataptr_solar[FRAMELEN-2]=FRAMELEN-12;
	dataptr_solar[FRAMELEN-1]=0;
        if(pattern==9 and length>1) dummy = (NFRAMES-iframe+iloop*remainder)%4;
      }
         
      if(pattern==5) // data for random mode
      {
	for(int z=0; z<FRAMELEN; z++) 
	{
	 dataptr_solar[z] = slidas_random();
      	}
      }
      
      if(pattern==0) // data for test mode
      {
        dataptr_solar[0]=1;
    	for(int z=1; z<FRAMELEN; z++)
	{
	  dataptr_solar[z]=2*dataptr_solar[z-1];
	}
      }

// print the data that solar sends out
	if (print==1)
	{
	  cout << "data sent:" << endl;
 	  for (int i=0; i<FRAMELEN; i++)
	  {
	    cout << dataptr_solar[i]<<", ";
	  }
	  cout << endl;
	}
    

// receiving each channel
      const bool choose1 = ((rand()%2)==0);
      int status;
      if(choose1) {
	for(int ir=0;ir<2;ir++) {
	  status = receive(ir+1,FRAMELEN,dataptr_solar, print);
	  if(status == -1) errcnt_frame[ir]+=1; else errcnt_word[ir]+=status;
	}
      } else {
	for(int ir=1;ir>=0;ir--) {
	  status = receive(ir+1,FRAMELEN,dataptr_solar, print);
	  if(status == -1) errcnt_frame[ir]+=1; else errcnt_word[ir]+=status;
	}
      }
      // re-arm receive FIFOs
      req_armed--;
      if(req_armed==0) {
	const bool choose2 = ((rand()%2)==0);
	if(choose2) {
	  for(int ir=0;ir<2;ir++) {
	    filar_setreq(FILAR_SLOT, ir+1, FILAR_ARM);
#ifdef ADD_SLEEP
	    if(ir==0) usleep(1000);
#endif
	  }
	} else {
	  for(int ir=1;ir>=0;ir--) {
	    filar_setreq(FILAR_SLOT, ir+1, FILAR_ARM);
#ifdef ADD_SLEEP
	    if(ir==1) usleep(500);
#endif
	  }
	}
	req_armed = FILAR_ARM;
      }
    // modify FRAMELEN for ROB block length running in pattern 9
    if (pattern==9 and length>1)
    {
      if (dummy==1) FRAMELEN=FRAMELENop-2;
      else if (dummy==2) FRAMELEN=FRAMELEN-FRAMELENop/4;
      else if (dummy==3) FRAMELEN=FRAMELENop-2;
      else FRAMELEN=FRAMELEN+FRAMELENop/4;
    }
   } // NFRAMES loop
  } // NUMLOOPS loop

  gettimeofday(&end_time, NULL);
  cout << "=========================================" << endl;
  cout << "FRAMELEN = " << FRAMELEN << endl;
  cout << "CHANNEL 0: frame errors = " << errcnt_frame[0] 
       << " , words errors = " << errcnt_word[0] << endl;
  cout << "CHANNEL 1: frame errors = " << errcnt_frame[1] 
       << " , words errors = " << errcnt_word[1] << endl;
  const double nframes = (1.0*NFRAMES)*(1.0*NUMLOOPS);
  const double nwords = nframes*(1.0*FRAMELEN);
  const double nbits = nwords*32.0;
  cout << "TOTAL: frames = " << scientific << nframes
       << ", words = " << nwords
       << ", bits = " << nbits << endl;
  //  const double dtime = end_time - beg_time;
  const double dtime = (end_time.tv_sec - beg_time.tv_sec) + (end_time.tv_usec - beg_time.tv_usec)/1000000.0;

  cout << "TOTAL TIME (sec) = " << dtime << endl;
  if(dtime>0) 
  {
    const double bps = 1.0*nbits/dtime;
    cout << "Bit/sec = " << bps << endl;
    cout << "Words/sec = " << nwords/dtime << endl;
    cout << "MB/sec = " << bps/(1024.0*1024.0*8.0) << endl;
    const double tproj = 1e15/bps;
    cout << "Time to test 1e15 bits: " << tproj << " sec, " 
	 << fixed << tproj/(60.0*60.0*24) << " days" << endl;
  }
  cout << "=========================================" << endl;
}


