/*
  Keeps sending random patterns throuugh LRL
  and verifying that they are received correctly.
*/
#include <stdio.h>
#include <cstdlib>
#include <cassert>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include <sys/time.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

//#define ADD_SLEEP

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 
#define FILAR_CHANNEL_0 0x1
#define FILAR_CHANNEL_1 0x2
#define FILAR_CHANNEL_MASK_0 0x1
#define FILAR_CHANNEL_MASK_1 0x2

#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

const char SEPARATOR[] = "=========================================================";

int str_to_int(string v) {
  int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

int receivedata(unsigned int FILAR_CHANNEL) {
    // use static array data[] for filar receive buffer:
    unsigned data[100000];
    // other variables
    int filsize = 0;
    int filerr(0);

    while (!filsize) {
      filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
    }

    cerr << "FILAR: received " << filsize 
	 << " words; error=" << hex << filerr << dec << endl;
    if(filerr!=0) {
      cerr << "ERROR: major frame error" << endl;
      exit(0);
    }
    int z;
    for(z=0; z<filsize; z++) {
      if(z>1 && z<filsize-2) continue; // for brevity
        cerr << "FILAR: data[" << z << "] = " 
            << hex << data[z] << dec << endl;
    }
    return(data[0]);
}

int main(int argc, char *argv[]) {
  int urlchannel = 2;
  int debugpattern=0;
  if(argc==1) {
    cerr << "USAGE: " << argv[0] << " CHANNEL_MASK IS_DEBUG" << endl;
    exit(0);
  }
  if(argc>=2) urlchannel = str_to_int(argv[1]);
  if(argc>=3) debugpattern = str_to_int(argv[1]);

  assert(urlchannel==FILAR_CHANNEL_MASK_0 || urlchannel==FILAR_CHANNEL_MASK_1 || urlchannel==FILAR_CHANNEL_MASK_BOTH);

  unsigned int *data_solar;
  
  cerr << "SOLAR: before solar_setup" << endl;
  solar_setup(0);
  solar_slidas_off(0);
  cerr << "SOLAR: after solar_setup" << endl;
  data_solar = solar_getbuffer(0);
  cerr << "SOLAR: retrieved solar buffer!"  << endl;
 
  // Set up two FILAR channels
  cerr << "FILAR: filar_setup with channel mask: " 
       << (FILAR_CHANNEL_MASK_BOTH) << endl;
  int setup_status = filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK_BOTH);
  if(setup_status==-1) {
    cerr << "ERROR: failed to bring up FILAR" << endl;
    exit(0);
  }
  
  // arm a few REQ FIFOs before going to event loo
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_1, FILAR_ARM);
 
  cerr << "STATS: done initialization" << endl;
  srand(time(0));

  unsigned int channel_mask = urlchannel;
  filar_seturl(0, channel_mask, 15);

  int i;
  const int url1[]={13, 1, 14, 2, 10, 3, 13, 4, 11, 5, 10, 6, 11, 7, 14, 8};
  const int url2[]={13, 1, 14, 2, 10, 3, 13, 4, 11, 5, 10, 9, 11, 7, 14, 8};
  for (i=0; i<16; i++){
    *(data_solar+0) = url1[i];
    solar_send(1,0);
    cerr << "Press ENTER to continue... ";
    cin.ignore();
    if(debugpattern) filar_seturl(0, channel_mask, url2[i]);
    else filar_seturl(0, channel_mask, url1[i]);
    for(unsigned int ich=1;ich<=2;ich++) {
      if(ich & channel_mask) {
	cerr << "RX(" << ich << "): " 
	     << hex << receivedata(ich) << endl;
      }
    }
    filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
    filar_setreq(FILAR_SLOT, FILAR_CHANNEL_1, FILAR_ARM);
  }

  cerr << "ALL DONE!" << endl;
}
