#include <stdio.h>
#include <cstdlib>
#include <cassert>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include <sys/time.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

//#define ADD_SLEEP

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 
#define FILAR_CHANNEL_0 0x1
#define FILAR_CHANNEL_1 0x2
#define FILAR_CHANNEL_MASK_0 0x1
#define FILAR_CHANNEL_MASK_1 0x2

#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

#define rdtscl(low) \
__asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")



#define BOF_WORD BOF0AAA0
#define EOF_WORD EOF0AAA0


int timing() {
    static __u32 told;
    __u32 tnew;
    int result;
    rdtscl(tnew);
    result = tstamp_to_us(told,tnew);
    told = tnew;
    return result;
}

int str_to_int(string v) {
  int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

int receivedata(unsigned int FILAR_CHANNEL) {
    // use static array data[] for filar receive buffer:
    unsigned data[100000];
    // other variables
    int filsize = 0;
    int filerr(0);
    int trial(0);

    while (!filsize & trial < 100) {
      filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
      trial++;
      }
    if (trial == 100) return(-1);

    cerr << "FILAR: received " << filsize 
	 << " words; error=" << hex << filerr << dec << endl;
    if(filerr!=0) {
      cerr << "ERROR: major frame error" << endl;
      exit(0);
      }
    int z;
    for(z=0; z<filsize; z++) {
      if(z>1 && z<filsize-2) continue; // for brevity
        cerr << "FILAR: data[" << z << "] = " 
            << hex << data[z] << dec << endl;
      }
  return(0);
}

int slidas_random_old(int reset)
{
  static unsigned reg_word = 0x524948e0; /* Random number seed */
  unsigned output_word;
  int i, j, xor_result;
  static int swap_map[32] = { 19,  0,  9, 20,  5, 31, 24, 22, 
			      1,  7, 27,  8, 17, 29, 23, 11, 
			      18, 21, 25,  2,  6, 13,  3, 10, 
			      30, 28, 14, 12,  4,  0, 26, 15};
  if (reset) reg_word = 0x524948e0;				   
  xor_result = 
    ((((reg_word & 0x80000000) && 1) ^     /* Bit 31 */
      ((reg_word & 0x20000000) && 1) ^     /* Bit 29 */
      ((reg_word & 0x40000) && 1)    ^     /* Bit 18 */
      ((reg_word & 0x20) && 1)) == 1);     /* Bit  5 */
  output_word = 0;
  for(j=0; j<32; j++) {
    if (j != 29) {
      if (j == 16)		 	    
	output_word ^= (xor_result << 16); 
      if (reg_word & (0x1 << j)) 	
	output_word ^= (0x1 << swap_map[j]);
    }
  }
  reg_word = output_word;
  return output_word; 
}

#define SEED 0x524948e0 /* Random number seed */

typedef union {
    unsigned int ival;
    unsigned short int sval[2];
} WORD;
    
static WORD reg_word;

unsigned int htab[65536], ltab[65536];

unsigned int slidas_random_next(unsigned int val)
{
    unsigned output_word, xor_result;
    int j;
    static int swap_map[32] = { 19,  0,  9, 20,  5, 31, 24, 22, 
				 1,  7, 27,  8, 17, 29, 23, 11, 
				18, 21, 25,  2,  6, 13,  3, 10, 
				30, 28, 14, 12,  4,  0, 26, 15 };
    

    xor_result =
          ((((val & 0x80000000) && 1) ^     /* Bit 31 */
            ((val & 0x20000000) && 1) ^     /* Bit 29 */
            ((val & 0x40000) && 1)    ^     /* Bit 18 */
            ((val & 0x20) && 1)) == 1);     /* Bit  5 */

    output_word = xor_result << 16;

    for(j=0; j<32; j++) {
	if (j==29)
	    continue;
	if (val & (0x1<<j)) 	
	    output_word ^= (0x1 << swap_map[j]);
    }
    return output_word; 
}


void slidas_initialize()
{
    WORD w;
    int i;

    for (i=0; i<65536; ++i) {
	w.sval[0] = i;
	w.sval[1] = 0;
	ltab[i] = slidas_random_next(w.ival);

	w.sval[0] = 0;
	w.sval[1] = i;
	htab[i] = slidas_random_next(w.ival);
    }
}

void slidas_reset()
{
  reg_word.ival = SEED;
}

unsigned int slidas_random()
{
  return reg_word.ival = htab[reg_word.sval[1]] ^ ltab[reg_word.sval[0]];
}

int receive(unsigned int FILAR_CHANNEL, int FRAMELEN, unsigned int *dataptr_solar)
{
  // use static array data[] for filar receive buffer:
  unsigned data[10000];

  // other variables
  int filsize = 0;
  int filerr(0);
  int errcnt = 0;
  while (!filsize) {
    filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
  }
  
  if(filerr!=0) {
    if(false) {
      cerr << "FILAR GROSS ERROR: received " << filsize
	   << " error=" << hex << filerr << dec << endl;
    }
    return -1;
  }
  assert(filsize==FRAMELEN);

  //  return 0; // FIXME

  for(int z=0; z<filsize; z++) {
    if(data[z] != dataptr_solar[z]) {
      errcnt+=1;
      if(false) {
	std::cerr << "FILAR ERROR: sent [" << hex << dataptr_solar[z] 
		  << "], received [" << data[z] << "]" << dec
		  << std::endl;
      }
    }
  }
  return errcnt;
}


/****************/
int reset_filar()
/****************/
{
  int channel;
  filar_cardreset(0);
  cerr << "Which channel would you like to reset?" << endl;
  cerr << "1 = Channel 1; 2 = Channel 2;" << endl;
  cin >> channel;
  filar_linkreset_hola(0, channel);
  return(0);
}

/****************/
int send_frames(int *Nframe_1, int *Nframe_2)
/****************/
{ 
  int FRAMELEN(254), NFRAMES;
  unsigned int *dataptr_solar;
  dataptr_solar = solar_getbuffer(0);
  // cerr << "Please set the length of each frame (126/254)" << endl;
  // cin >> FRAMELEN;
  cerr << "Please input the number of frames you would like to send" << endl;
  cerr << "(Be careful not to exceed a total of 19 frames at each channel.)" << endl;
  cin >> NFRAMES;
  if (*Nframe_1+NFRAMES > 19 | *Nframe_2+NFRAMES >19) {
    cerr << "Warning: the total number of frames will exceed 19..." << endl;
    if (*Nframe_1 > *Nframe_2) NFRAMES = 19 - *Nframe_1;
    // else NFRAMES = 19 - *Nframe_2;
    cerr << "Send " << NFRAMES << " frames instead..." << endl;
    }

  // prepare some dummy data
  int sent=0;
  for(int il=0;il<FRAMELEN;il++) {*(dataptr_solar+il)=0xaaaaaaaa+il;}
/*
  cerr << "SOLAR: sending " << FRAMELEN << " words" << endl;
  solar_send(FRAMELEN,0); 
  sent++;
  cerr << "SOLAR: sent " << FRAMELEN << " words" << endl;
  cerr << "---> sent count = " << sent << endl;
 */ 
  // now try to trigger XOFF from FILAR end
  cerr << "SOLAR: sending data train of " << NFRAMES << " frames" << endl;
  for(int il=0;il<NFRAMES;il++) {
    *(dataptr_solar+FRAMELEN-1) = 0xabcdaaaa+il;
    solar_send(FRAMELEN,0); sent++;
    cerr << "---> sent count = " << sent << endl;
    }

  // Continuous loop to control data reception
  cerr << "FYI, sent a total of " << sent << " frames ("
       << sent*FRAMELEN << ") words" << endl;
  *Nframe_1 = *Nframe_1+NFRAMES;
  *Nframe_2 = *Nframe_2+NFRAMES;
  filar_retbuf(0,0,1);
  filar_retbuf(0,1,1);
  return(sent);
}

/****************/
int receive_frames(int *Nframe_1, int *Nframe_2)
/****************/
{
  int chanmask, nreceive;
  int req_armed[] = {FILAR_ARM, FILAR_ARM};
  int totr[] = {1,1};
  int frame_index;
  cerr << "Which channel would you like to receive data from?" << endl;
  cerr << "1 = Channel 1; 2 = Channel 2;" << endl;
  cin >> chanmask;
  cerr << "How many frames do you want to receive?" << endl;
  cin >> nreceive;
  cerr << "User input: " << chanmask << " " << nreceive << endl;
  //receiving data
  for(int ir=0;ir<nreceive;ir++) {
    if(chanmask == 1) {
      cerr << "=========== " << totr[0] << " - CHANNEL" 
      << FILAR_CHANNEL_0 << "===========" << endl;
      frame_index = receivedata(FILAR_CHANNEL_0);
      req_armed[0]--;
      if (frame_index < 0) break;
      totr[0]++;
      }
    if(chanmask == 2) {
      cerr << "=========== " << totr[1] << " - CHANNEL" 
      << FILAR_CHANNEL_1 << "===========" << endl;
      frame_index = receivedata(FILAR_CHANNEL_1);
      req_armed[1]--;
      if (frame_index < 0) break;
      totr[1]++;
      }
      // time to re-arm the REQ FIFO
    for(unsigned int ichan=0; ichan<=1; ichan++) {
      if(req_armed[ichan]==0) {
	cerr << "INFO: Arming " << ichan << " with " << FILAR_ARM << " REQ FIFO buffers..." << endl;
	filar_setreq(FILAR_SLOT, ichan==0 ? FILAR_CHANNEL_0 : FILAR_CHANNEL_1, FILAR_ARM);
	sleep(0.1);
	req_armed[ichan] = FILAR_ARM;
	cerr << "Armed!" << endl;
	}
      }
      // if (frame_index < 0) break;
    }
  if (chanmask == 1) *Nframe_1 = *Nframe_1-totr[0]+1;
  else *Nframe_2 = *Nframe_2-totr[1]+1;
  // cout << frame_index << " " << totr[0] << " " << totr[1] << endl;
  filar_retbuf(0,0,1);
  filar_retbuf(0,1,1);
  return(0);
}

/****************/
int set_url()
/****************/
{
  int channel;
  int pattern;
  cerr << "Which channel would you like to send the URL line through?" << endl;
  cerr << "1 = Channel 1; 2 = Channel 2;" << endl;
  cin >> channel;
  cerr << "Do you want to send a specific number or just the flow control activation pattern?" << endl;
  cerr << "(Input an integer between 0 and 15, or input 99 to send the activation pattern.)" << endl;
  cin >> pattern;
  if (pattern >=0 & pattern <16) filar_seturl(0, channel, pattern);
  else if (pattern == 99) filar_flowctrlenable(0, channel);
  else {
    cerr << "Input error. Back to the main menu." << endl;
    return(0);
    }
  cerr << "URL line has been sent." << endl;
  return(0);
}

/****************/
int print_lrl()
/****************/
{
  int pattern;
  pattern = solar_register(0);
  cerr << "The LRL line is " << pattern << endl;
  return(0);
}

/***********/
int main() 
/***********/
{
  int choice;
  int setup_status;
  int frame_index, Nframe_1(0), Nframe_2(0);
  cerr << "SOLAR: before solar_setup" << endl;
  solar_setup(0);
  solar_slidas_off(0);
  cerr << "SOLAR[" << "]: after solar_setup" << endl;
  //  dataptr_solar = solar_getbuffer(0);
    cerr << "SOLAR: retrieved solar buffer!"  << endl;

  // Set up two FILAR channels
  cerr << "FILAR: filar_setup with channel mask: " 
       << (FILAR_CHANNEL_MASK_BOTH) << endl;
  setup_status = filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK_BOTH);
  if(setup_status==-1) {
    cerr << "ERROR: failed to bring up FILAR" << endl;
    exit(0);
  }
  // arm a few REQ FIFOs before going to event loop
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_1, FILAR_ARM);
  int req_armed = FILAR_ARM;

  while (choice !=0) {
  cerr << "===============================================================================" << endl;
  cerr << "This is the hola test program. Please select an option." << endl;
  cerr << "0 = exit;" << endl;
  cerr << "1 = reset FILAR;" << endl;
  cerr << "2 = sends frames through the forward channel;" << endl;
  cerr << "3 = receives frames through FILAR channel;" << endl;
  cerr << "4 = set channel URL;" << endl;
  cerr << "5 = print the LRL from the DAQ channel;" << endl;
  cerr << "Reminder: You have " << endl;
  cerr << Nframe_1 << " frame(s) at Channel 1 and " << endl; 
  cerr << Nframe_2 << " frame(s) at Channel 2 waiting..." << endl;
  cerr << "===============================================================================" << endl;
  cin >> choice;
  if (choice == 0) exit(0);
  else if (choice == 1) reset_filar();
  else if (choice == 2) send_frames(&Nframe_1, &Nframe_2);
  else if (choice == 3) receive_frames(&Nframe_1, &Nframe_2);
  else if (choice == 4) set_url();
  else if (choice == 5) print_lrl();
  else cerr << "Input error. Please reselect your option." << endl;
  }
}

