#include <stdio.h>
#include <cstdlib>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

#define BOF_WORD BOF0AAA0
#define EOF_WORD EOF0AAA0

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 

// first FILAR channel:
#if 0
#define FILAR_CHANNEL_MASK 0x1
#define FILAR_CHANNEL_0 0x1
#else
// OR: second FILAR channel:
#define FILAR_CHANNEL_MASK 0x2
#define FILAR_CHANNEL_0 0x2
#endif
#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

#define rdtscl(low) \
__asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")

int timing() {
    static __u32 told;
    __u32 tnew;
    int result;
    rdtscl(tnew);
    result = tstamp_to_us(told,tnew);
    told = tnew;
    return result;
}

int str_to_int(string v) {
  int result;
  istringstream buf(v);
  buf >> result;
  return result;
}

int slidas_random_old(int reset)
{
  static unsigned reg_word = 0x524948e0; /* Random number seed */
  unsigned output_word;
  int i, j, xor_result;
  static int swap_map[32] = { 19,  0,  9, 20,  5, 31, 24, 22, 
			      1,  7, 27,  8, 17, 29, 23, 11, 
			      18, 21, 25,  2,  6, 13,  3, 10, 
			      30, 28, 14, 12,  4,  0, 26, 15};
  if (reset) reg_word = 0x524948e0;				   
  xor_result = 
    ((((reg_word & 0x80000000) && 1) ^     /* Bit 31 */
      ((reg_word & 0x20000000) && 1) ^     /* Bit 29 */
      ((reg_word & 0x40000) && 1)    ^     /* Bit 18 */
      ((reg_word & 0x20) && 1)) == 1);     /* Bit  5 */
  output_word = 0;
  for(j=0; j<32; j++) {
    if (j != 29) {
      if (j == 16)		 	    
	output_word ^= (xor_result << 16); 
      if (reg_word & (0x1 << j)) 	
	output_word ^= (0x1 << swap_map[j]);
    }
  }
  reg_word = output_word;
  return output_word; 
}

#define SEED 0x524948e0 /* Random number seed */

typedef union {
    unsigned int ival;
    unsigned short int sval[2];
} WORD;
    
static WORD reg_word;

unsigned int htab[65536], ltab[65536];

unsigned int slidas_random_next(unsigned int val)
{
    unsigned output_word, xor_result;
    int j;
    static int swap_map[32] = { 19,  0,  9, 20,  5, 31, 24, 22, 
				 1,  7, 27,  8, 17, 29, 23, 11, 
				18, 21, 25,  2,  6, 13,  3, 10, 
				30, 28, 14, 12,  4,  0, 26, 15 };
    

    xor_result =
          ((((val & 0x80000000) && 1) ^     /* Bit 31 */
            ((val & 0x20000000) && 1) ^     /* Bit 29 */
            ((val & 0x40000) && 1)    ^     /* Bit 18 */
            ((val & 0x20) && 1)) == 1);     /* Bit  5 */

    output_word = xor_result << 16;

    for(j=0; j<32; j++) {
	if (j==29)
	    continue;
	if (val & (0x1<<j)) 	
	    output_word ^= (0x1 << swap_map[j]);
    }
    return output_word; 
}


void slidas_initialize()
{
    WORD w;
    int i;

    for (i=0; i<65536; ++i) {
	w.sval[0] = i;
	w.sval[1] = 0;
	ltab[i] = slidas_random_next(w.ival);

	w.sval[0] = 0;
	w.sval[1] = i;
	htab[i] = slidas_random_next(w.ival);
    }
}

void slidas_reset()
{
    reg_word.ival = SEED;
}

unsigned int slidas_random()
{
    return reg_word.ival = htab[reg_word.sval[1]] ^ ltab[reg_word.sval[0]];
}



#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
int receive(unsigned int FILAR_CHANNEL, unsigned int *rcv)
{
  // use static array data[] for filar receive buffer:
  unsigned data[10000];
  // what solar sent out (via SLIDAS mode)
  unsigned int dataptr_solar[10000];

  // other variables
  int filsize = 0;
  int filerr(0);
  int errcnt = 0;
  while (!filsize) {
    filsize=filar_receive_channel_ptr( FILAR_SLOT, &filerr, FILAR_CHANNEL, (unsigned*)(&data));
  }
  *rcv = filsize;

  if(filerr!=0) {
    cerr << "FILAR ERROR: received " << filsize
	 << " error=" << hex << filerr << dec << endl;
    return -1;
  }

  // return 0; // FIXME

  // pre-fill the SLIDAS check buffer
  // Note: this is the most time-consuming part
  // and we should think about moving it to a separate thread.
  for(int z=0; z<filsize; z++) {
    dataptr_solar[z] = slidas_random();
  }

  for(int z=0; z<filsize; z++) {
    if(data[z] != dataptr_solar[z]) {
      errcnt+=1;
      std::cerr << "FILAR ERROR: sent [" << hex << dataptr_solar[z] 
		<< "], received [" << data[z] << "]" << dec
		<< std::endl;
    }
  }
  return errcnt;
}

int main( int argc, char* argv[]) {
  unsigned int pattern(5), mode(0), length(4);
  long int NFRAMES(3000000);
  long int FRAMELEN(254);
  if(argc==1) {
    printf("Select the pattern switch (SW1) setting:\n");
    printf(" 0 = Testmode 1 = Walking 1\n");
    printf(" 2 = Walking 0 3 = FFFF/0000\n");
    printf(" 4 = AAAA/5555 5 = Random data\n");
    printf(" 6 = Random data, no control 7 = Random control\n");
    printf(" 8 = ROB blocks, 1 lenght 9 = ROB blocks, 3 lengths\n");
    printf("=============================================================\n");
    printf("Select the mode switch (SW2) setting:\n");
    printf(" 0 = RUN 1 = Single word\n");
    printf(" 2 = Single block 3 = 50 kHz\n");
    printf(" 4 = 75 kHz 5 = 90 kHz\n");
    printf(" 6 = 100 kHz 7 = 110 kHz\n");
    printf(" 8 = 125 kHz 9 = 150 kHz\n");
    printf("=============================================================\n");
    printf("Select the length switch (SW3) setting:\n");
    printf(" 0 = 52 byte (header only) 1 = 64 byte\n");
    printf(" 2 = 256 byte 3 = 512 byte\n");
    printf(" 4 = 1 kbyte 5 = 2 kbyte\n");
    printf(" 6 = 4 kbyte 7 = 8 kbyte\n");
    printf(" 8 = 16 kbyte 9 = 32 kbyte\n");
    cout << "USAGE: " << argv[0] << " NFRAMES pattern mode length" << endl;
    exit(0);
  }
  NFRAMES = str_to_int(argv[1]);
  //pattern = str_to_int(argv[2]);
  //mode = str_to_int(argv[3]);
  //length = str_to_int(argv[4]);
  std::cerr << "pattern = " << pattern << " | mode = " << mode 
	    << " | length = " << length << std::endl;
  std::cerr << "NFRAMES=" << NFRAMES << "; FRAMELEN="
	    << FRAMELEN << std::endl;

  reg_word.ival = SEED;
  slidas_initialize();
  
  unsigned int *dataptr_solar;

  cerr << "SOLAR: before solar_setup" << endl;
  solar_setup(0);
  solar_slidas_off(0);
  cerr << "SOLAR[" << "]: after solar_setup" << endl;
  dataptr_solar = solar_getbuffer(0);
  cerr << "SOLAR: retrieved solar buffer!"  << endl;

  // We have a faulty SOLAR that doesn't always come up
  // The way to detect failure is by looking at FILAR LDOWN
  // If detected, we re-reset SOLAR and try again.
  int linkup=0;
  for(int itry=0; itry<3; itry++) {
    cerr << "FILAR: before filar_setup (attempt " << itry << ")" << endl;   
    if( filar_setup_hola(FILAR_SLOT, FILAR_CHANNEL_MASK) == -1) {
      solar_setup(0);
      cerr << "SOLAR[" << "]: after solar_setup" << endl;
      dataptr_solar = solar_getbuffer(0);
    }
    else {
      cerr << "FILAR[" << "] after filar_setup" << endl;
      linkup=1;
      break;
    }
  }
  if(linkup==0) {
    cerr << "ERROR: unable to bring up FILAR" << endl;
    exit(0);
  }
  // arm a few REQ FIFOs before going to event loop
  filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
  int req_armed = FILAR_ARM;

  // enter slidas mode
  solar_slidas_config(0,pattern,length,mode);
  solar_slidas_dump(0);
  
  cerr << "Running tests. You won't see any output until it's done..." << endl;

  // prepare some dummy data
  long int iframe = NFRAMES;
  long int errcnt_word = 0;
  long int errcnt_frame = 0;
  time_t beg_time = time(NULL);
  while(iframe-- > 0) {
    if(iframe%1000000==0) {
      cout << iframe/1000000 << " 1M groups left" << endl;
    }
    // receiving
    for(int ir=0;ir<1;ir++) {
      unsigned int rcvd;
      int status = receive(FILAR_CHANNEL_0,&rcvd);
      FRAMELEN = rcvd;
      if(status == -1) errcnt_frame++; else errcnt_word+=status;
      req_armed--;
      // time to re-arm the REQ FIFO
      if(req_armed<2) {
	filar_setreq(FILAR_SLOT, FILAR_CHANNEL_0, FILAR_ARM);
	//sleep(0.1);
	req_armed = FILAR_ARM;
      }
    }
  }
  time_t end_time = time(NULL);
  cout << "=========================================" << endl;
  cout << "FRAMELEN = " << FRAMELEN << endl;
  cout << "BAD: frames = " << errcnt_frame 
       << ", words = " << errcnt_word << endl;
  const double nwords = (1.0*NFRAMES)*FRAMELEN;
  const double nbits = (1.0*NFRAMES)*(FRAMELEN*32.0);
  cout << "TOTAL: frames = " << NFRAMES 
       << ", words = " << nwords
       << ", bits = " << nbits << endl;
  const double dtime = end_time - beg_time;
  cout << "TOTAL TIME (sec) = " << dtime << endl;
  if(dtime>0) {
    const double bps = 1.0*nbits/dtime;
    cout << "Bit/sec = " << bps << endl;
    cout << "MBit/sec = " << bps/(1024.0*1024.0) << endl;
    cout << "MB/sec = " << bps/(1024.0*1024.0*8.0) << endl;
    const double tproj = 1e15/bps;
    cout << "Time to test 1e15 bits: " << tproj << " sec, " << tproj/(60.0*60.0*24) << " days" << endl;
  }
  cout << "=========================================" << endl;
}


