#include <stdio.h>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <cctype>
#include <string>
#include <unistd.h>
#include <sched.h>
#include <linux/types.h>
#include "NodeUtils.hh"
using namespace std;

// s32pci64 headers
extern "C" {
#include "s32pci64-filar.h"
#include "s32pci64-solar.h"
}

// Number of REQ FIFOs to arm in the beginning
#define FILAR_ARM 15
// PCI slot in which FILAR is installed
#define FILAR_SLOT 0 

#define FILAR_SLOT 0 
#define FILAR_CHANNEL_0 0x1
#define FILAR_CHANNEL_1 0x2
#define FILAR_CHANNEL_MASK_0 0x1
#define FILAR_CHANNEL_MASK_1 0x2

#define FILAR_CHANNEL_MASK_BOTH 0x1|0x2

int str_to_int(string v) {
  int result;
  istringstream buf(v);
  buf >> result;
  return result;
}
void capitalize(string& v) {
  for(unsigned int i=0;i<v.length();i++) {
    v[i] = tolower(v[i]);
  }
}

/* Non-blocking function that tries to receive one S-Link frame */
int receive(unsigned int FILAR_CHANNEL, unsigned int *data, int print) {
  int nattempts = 10;
  // other variables
  int filsize = 0;
  int filerr = 0;
  int errcnt = 0;
  while (!filsize && nattempts-->0) {
    filsize=filar_receive_channel_ptr(FILAR_SLOT,&filerr,FILAR_CHANNEL,data);
  }
  
  // print received data
  if (print==1) {
    cout <<"data received:"<<endl;
    if(filsize<50) {
      for (int i=0; i<filsize; i++) {
	cout<< i << " : " << "0x" << hex << data[i] << dec << endl;
      } 
    }
    else {
      for (int i=0; i<25; i++) {
	cout<< i << " : " << "0x" << hex << data[i] << dec << endl;
      }
      cout << "............" << endl;
      for (int i=filsize-25; i<filsize; i++) {
	cout<< i << " : " << "0x" << hex << data[i] << dec << endl;
      }
    }
    cout<<endl;
  }
  
  if(filerr!=0) {
    cerr << "FILAR ERROR: received " << filsize << " words"
	 << " with error status " << hex << filerr << dec << endl;
    return -1;
  }

  return filsize;
}

void print_help() {
  cerr << "SOLAR COMMANDS:" << endl;
  cerr << 
    "  si=solar_init | sc=solar_cardreset | sl=solar_linkreset" << endl;
  cerr <<
    "  son=solar_ureset_on | soff=solar_ureset_off" << endl;
  cerr << "FILAR COMMANDS:" << endl;
  cerr <<
    "  fi=filar_init | fc=filar_cardreset | fl=filar_linkreset <CHAN_MASK>" << endl;
  cerr << "  fon=filar_ureset_on | foff=filar_ureset_off" << endl;
  cerr << "SHORTCUT COMMANDS:" << endl;
  cerr <<
    "  default=si+fi+sc+fc" << endl;
}

int main( int argc, char* argv[]) {
  string input,com;
  int arg1,arg2;
  unsigned int data[10000]; //filar receive buffer
  unsigned int *dataptr_solar;
  int chan=FILAR_CHANNEL_1; //default FILAR channel

  // command line parser
  print_help();

  while(true) {
    com="";
    arg1=-1;
    arg2=-1;
    cout << "SOLAR_FILAR > ";
    getline(cin, input);
    istringstream iss(input);
    iss >> com;
    iss >> arg1;
    iss >> arg2;
    capitalize(com);
    if(com=="exit" || com=="quit" || com=="q" || com==".q") {
      exit(0);
    }
    else if(com=="chan") {
      if(arg1==-1) {
	cout << "Using channel: " << chan << endl;
      }
      else {
	if(arg1==FILAR_CHANNEL_MASK_0 || arg1==FILAR_CHANNEL_MASK_1) {
	  chan = arg1;
	} else {
	  cerr << "Bad channel given:"
	       << arg1 << "; keeping old channel: " << chan << endl;
	  continue;
	}
	cout << "Switched to channel: " << chan << endl;
      }
    }
    else if(com=="default") {
      // shortcut to set everything up
      solar_ak0_init(0);
      filar_ak0_init(0);
      solar_ak1_cardreset(0);
      filar_ak1_cardreset(0);
    }
    else if(com=="solar_init" || com=="si") {
      solar_ak0_init(0);
    }
    else if(com=="solar_cardreset" || com=="sc") {
      solar_ak1_cardreset(0);
    }
    else if(com=="solar_ureset_on" || com=="son") {
      solar_ak_ureset_on(0);
    }    
    else if(com=="s_ureset_off" || com=="soff") {
      solar_ak_ureset_off(0);
    }
    else if(com=="send") {
      *(dataptr_solar+0) = 0;
      *(dataptr_solar+1) = 0;
      if(arg1!=-1) {
	*(dataptr_solar+0) = arg1;
	*(dataptr_solar+1) = arg1;
      }
      solar_send(1,0);
    }
    else if(com=="solar_linkreset" || com=="sl") {
      solar_ak2_linkreset(0);
      dataptr_solar = solar_getbuffer(0);
    }
    else if(com=="filar_init" || com=="fi") {
      filar_ak0_init(0);
    }
    else if(com=="filar_cardreset" || com=="fc") {
      filar_ak1_cardreset(0);
    }
    else if(com=="filar_ureset_on" || com=="fon") {
      filar_ak_ureset_on(0);
    }    
    else if(com=="filar_ureset_off" || com=="foff") {
      filar_ak_ureset_off(0);
    }
    else if(com=="filar_linkreset" || com=="fl") {
      filar_ak2_linkreset(0,chan);
    }
    else if(com=="arm") {
      filar_setreq(FILAR_SLOT, chan, FILAR_ARM);
      cout << "Rearmed FILAR channel " << chan 
	   << " with " << FILAR_ARM << " buffers" << endl;
    }
    else if(com=="xon") {
      filar_flowctrlenable(0, chan); 
    }
    else if(com=="xoff") {
      filar_flowctrldisable(0, chan); 
    }
    else if(com=="get") {
      int nwords = receive(chan,data,0);
      cout << "Received: " << nwords << " words" << endl;
    }
    else if(com=="dump") {
      int nwords = receive(chan,data,1);
      cout << "Received: " << nwords << " words" << endl;
    }
    else {
      print_help();
      continue;
    }
  }
  
}
